# Inspiration: https://github.com/apify/crawlee/blob/v3.7.3/packages/core/src/autoscaling/snapshotter.ts

from __future__ import annotations

from datetime import datetime, timedelta, timezone
from logging import getLogger
from typing import TYPE_CHECKING, cast

import psutil

from crawlee._utils.byte_size import ByteSize
from crawlee._utils.recurring_task import RecurringTask
from crawlee.autoscaling.types import ClientSnapshot, CpuSnapshot, EventLoopSnapshot, MemorySnapshot, Snapshot
from crawlee.events.types import Event, EventSystemInfoData

if TYPE_CHECKING:
    from types import TracebackType

    from crawlee.events import EventManager

logger = getLogger(__name__)


class Snapshotter:
    """Monitors and logs system resource usage at predefined intervals for performance optimization.

    The class monitors and records the state of various system resources (CPU, memory, event loop, and client API)
    at predefined intervals. This continuous monitoring helps in identifying resource overloads and ensuring optimal
    performance of the application. It is utilized in the `AutoscaledPool` module to adjust task allocation
    dynamically based on the current demand and system load.
    """

    def __init__(
        self,
        event_manager: EventManager,
        *,
        event_loop_snapshot_interval: timedelta = timedelta(milliseconds=500),
        client_snapshot_interval: timedelta = timedelta(milliseconds=1000),
        max_used_cpu_ratio: float = 0.95,
        max_memory_size: ByteSize | None = None,
        max_used_memory_ratio: float = 0.7,
        max_event_loop_delay: timedelta = timedelta(milliseconds=50),
        max_client_errors: int = 1,
        snapshot_history: timedelta = timedelta(seconds=30),
        reserve_memory_ratio: float = 0.5,
        memory_warning_cooldown_period: timedelta = timedelta(milliseconds=10000),
        client_rate_limit_error_retry_count: int = 2,
    ) -> None:
        """Creates a new instance.

        Args:
            event_manager: The event manager used to emit system info events. From data provided by this event
              the CPU and memory usage are read.

            event_loop_snapshot_interval: The interval at which the event loop is sampled.

            client_snapshot_interval: The interval at which the client is sampled.

            max_used_cpu_ratio: Sets the ratio, defining the maximum CPU usage. When the CPU usage is higher than
                the provided ratio, the CPU is considered overloaded.

            max_memory_size: Sets the maximum amount of system memory to be used by the `AutoscaledPool`. If `None`
                is provided, the max amount of memory to be used is set to one quarter of total system memory.
                I.e. on a system with 8192 MB, the `AutoscaledPool` will only use up to 2048 MB of memory.

            max_used_memory_ratio: Sets the ratio, defining the maximum ratio of memory usage. When the memory usage
                is higher than the provided ratio of `max_memory_size`, the memory is considered overloaded.

            max_event_loop_delay: Sets the maximum delay of the event loop. When the delay is higher than the provided
                value, the event loop is considered overloaded.

            max_client_errors: Sets the maximum number of client errors (HTTP 429). When the number of client errors
                is higher than the provided number, the client is considered overloaded.

            snapshot_history: Sets the time interval for which the snapshots are kept.

            reserve_memory_ratio: Fraction of memory kept in reserve. Used to calculate critical memory overload
                threshold.

            memory_warning_cooldown_period: Minimum time interval between logging successive critical memory overload
                warnings.

            client_rate_limit_error_retry_count: Number of retries for a client request before considering it a failure
                due to rate limiting.
        """
        self._event_manager = event_manager
        self._event_loop_snapshot_interval = event_loop_snapshot_interval
        self._client_snapshot_interval = client_snapshot_interval
        self._max_event_loop_delay = max_event_loop_delay
        self._max_used_cpu_ratio = max_used_cpu_ratio
        self._max_used_memory_ratio = max_used_memory_ratio
        self._max_client_errors = max_client_errors
        self._snapshot_history = snapshot_history
        self._reserve_memory_ratio = reserve_memory_ratio
        self._memory_warning_cooldown_period = memory_warning_cooldown_period
        self._client_rate_limit_error_retry_count = client_rate_limit_error_retry_count
        self._max_memory_size = max_memory_size or self._get_default_max_memory_size()

        self._cpu_snapshots: list[CpuSnapshot] = []
        self._event_loop_snapshots: list[EventLoopSnapshot] = []
        self._memory_snapshots: list[MemorySnapshot] = []
        self._client_snapshots: list[ClientSnapshot] = []

        self._snapshot_event_loop_task = RecurringTask(self._snapshot_event_loop, self._event_loop_snapshot_interval)
        self._snapshot_client_task = RecurringTask(self._snapshot_client, self._client_snapshot_interval)

        self._timestamp_of_last_memory_warning: datetime = datetime.now(timezone.utc) - timedelta(hours=1)

    @staticmethod
    def _get_default_max_memory_size() -> ByteSize:
        """Default `memory_max_size` is 1/4 of the total system memory."""
        max_memory_size_in_bytes = int(psutil.virtual_memory().total * 0.25)
        max_memory_size = ByteSize(max_memory_size_in_bytes)
        logger.info(f'Setting max_memory_size of this run to {max_memory_size}.')
        return max_memory_size

    async def __aenter__(self) -> Snapshotter:
        """Starts capturing snapshots at configured intervals."""
        self._event_manager.on(event=Event.SYSTEM_INFO, listener=self._snapshot_cpu)
        self._event_manager.on(event=Event.SYSTEM_INFO, listener=self._snapshot_memory)
        self._snapshot_event_loop_task.start()
        self._snapshot_client_task.start()
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        """Stops all resource capturing.

        This method stops capturing snapshots of system resources (CPU, memory, event loop, and client information).
        It should be called to terminate resource capturing when it is no longer needed.
        """
        self._event_manager.off(event=Event.SYSTEM_INFO, listener=self._snapshot_cpu)
        self._event_manager.off(event=Event.SYSTEM_INFO, listener=self._snapshot_memory)
        await self._snapshot_event_loop_task.stop()
        await self._snapshot_client_task.stop()

    def get_memory_sample(self, duration: timedelta | None = None) -> list[Snapshot]:
        """Returns a sample of the latest memory snapshots.

        Args:
            duration: The duration of the sample from the latest snapshot. If omitted, it returns a full history.

        Returns:
            A sample of memory snapshots.
        """
        snapshots = cast(list[Snapshot], self._memory_snapshots)
        return self._get_sample(snapshots, duration)

    def get_event_loop_sample(self, duration: timedelta | None = None) -> list[Snapshot]:
        """Returns a sample of the latest event loop snapshots.

        Args:
            duration: The duration of the sample from the latest snapshot. If omitted, it returns a full history.

        Returns:
            A sample of event loop snapshots.
        """
        snapshots = cast(list[Snapshot], self._event_loop_snapshots)
        return self._get_sample(snapshots, duration)

    def get_cpu_sample(self, duration: timedelta | None = None) -> list[Snapshot]:
        """Returns a sample of the latest CPU snapshots.

        Args:
            duration: The duration of the sample from the latest snapshot. If omitted, it returns a full history.

        Returns:
            A sample of CPU snapshots.
        """
        snapshots = cast(list[Snapshot], self._cpu_snapshots)
        return self._get_sample(snapshots, duration)

    def get_client_sample(self, duration: timedelta | None = None) -> list[Snapshot]:
        """Returns a sample of the latest client snapshots.

        Args:
            duration: The duration of the sample from the latest snapshot. If omitted, it returns a full history.

        Returns:
            A sample of client snapshots.
        """
        snapshots = cast(list[Snapshot], self._client_snapshots)
        return self._get_sample(snapshots, duration)

    @staticmethod
    def _get_sample(snapshots: list[Snapshot], duration: timedelta | None = None) -> list[Snapshot]:
        """Returns a time-limited sample from snapshots or full history if duration is None."""
        if not duration:
            return snapshots

        if not snapshots:
            return []

        latest_time = snapshots[-1].created_at
        return [snapshot for snapshot in reversed(snapshots) if latest_time - snapshot.created_at <= duration]

    def _snapshot_cpu(self, event_data: EventSystemInfoData) -> None:
        """Captures a snapshot of the current CPU usage.

        This method does not perform CPU usage measurement. Instead, it just reads the data received through
        the `event_data` parameter, which is expected to be supplied by the event manager.

        Args:
            event_data: System info data from which CPU usage is read.
        """
        snapshot = CpuSnapshot(
            used_ratio=event_data.cpu_info.used_ratio,
            max_used_ratio=self._max_used_cpu_ratio,
            created_at=event_data.cpu_info.created_at,
        )

        snapshots = cast(list[Snapshot], self._cpu_snapshots)
        self._prune_snapshots(snapshots, event_data.cpu_info.created_at)
        self._cpu_snapshots.append(snapshot)

    def _snapshot_memory(self, event_data: EventSystemInfoData) -> None:
        """Captures a snapshot of the current memory usage.

        This method does not perform memory usage measurement. Instead, it just reads the data received through
        the `event_data` parameter, which is expected to be supplied by the event manager.

        Args:
            event_data: System info data from which memory usage is read.
        """
        snapshot = MemorySnapshot(
            total_size=event_data.memory_info.total_size,
            current_size=event_data.memory_info.current_size,
            max_memory_size=self._max_memory_size,
            max_used_memory_ratio=self._max_used_memory_ratio,
            created_at=event_data.memory_info.created_at,
        )

        snapshots = cast(list[Snapshot], self._memory_snapshots)
        self._prune_snapshots(snapshots, snapshot.created_at)
        self._memory_snapshots.append(snapshot)

        self._evaluate_memory_load(event_data.memory_info.current_size, event_data.memory_info.created_at)

    def _snapshot_event_loop(self) -> None:
        """Captures a snapshot of the current event loop usage.

        This method evaluates the event loop's latency by comparing the expected time between snapshots to the actual
        time elapsed since the last snapshot. The delay in the snapshot reflects the time deviation due to event loop
        overhead - it's calculated by subtracting the expected interval between snapshots from the actual time elapsed
        since the last snapshot. If there's no previous snapshot, the delay is considered zero.
        """
        snapshot = EventLoopSnapshot(max_delay=self._max_event_loop_delay, delay=timedelta(seconds=0))
        previous_snapshot = self._event_loop_snapshots[-1] if self._event_loop_snapshots else None

        if previous_snapshot:
            event_loop_delay = snapshot.created_at - previous_snapshot.created_at - self._event_loop_snapshot_interval
            snapshot.delay = event_loop_delay

        snapshots = cast(list[Snapshot], self._event_loop_snapshots)
        self._prune_snapshots(snapshots, snapshot.created_at)
        self._event_loop_snapshots.append(snapshot)

    def _snapshot_client(self) -> None:
        """Captures a snapshot of the current API state by checking for rate limit errors (HTTP 429).

        Only errors produced by a 2nd retry of the API call are considered for snapshotting since earlier errors may
        just be caused by a random spike in the number of requests and do not necessarily signify API overloading.
        """
        # TODO: This is just a dummy placeholder. It can be implemented once `StorageClient` is ready.
        # Attribute `self._client_rate_limit_error_retry_count` will be used here.
        # https://github.com/apify/crawlee-py/issues/60

        error_count = 0
        snapshot = ClientSnapshot(error_count=error_count, max_error_count=self._max_client_errors)

        snapshots = cast(list[Snapshot], self._client_snapshots)
        self._prune_snapshots(snapshots, snapshot.created_at)
        self._client_snapshots.append(snapshot)

    def _prune_snapshots(self, snapshots: list[Snapshot], now: datetime) -> None:
        """Removes snapshots that are older than the `self._snapshot_history`.

        This method modifies the list of snapshots in place, removing all snapshots that are older than the defined
        snapshot history relative to the `now` parameter.

        Args:
            snapshots: List of snapshots to be pruned in place.
            now: The current date and time, used as the reference for pruning.
        """
        # Find the index where snapshots start to be within the allowed history window.
        # We'll keep snapshots from this index onwards.
        keep_from_index = None
        for i, snapshot in enumerate(snapshots):
            if now - snapshot.created_at <= self._snapshot_history:
                keep_from_index = i
                break

        # If all snapshots are old, keep_from_index will remain None, so we clear the list.
        # Otherwise, we keep only the recent snapshots.
        if keep_from_index is not None:
            del snapshots[:keep_from_index]
        else:
            snapshots.clear()

    def _evaluate_memory_load(self, current_memory_usage_size: ByteSize, snapshot_timestamp: datetime) -> None:
        """Evaluates and logs critical memory load conditions based on the system information.

        Args:
            current_memory_usage_size: The current memory usage.
            snapshot_timestamp: The time at which the memory snapshot was taken.
        """
        # Check if the warning has been logged recently to avoid spamming
        if snapshot_timestamp < self._timestamp_of_last_memory_warning + self._memory_warning_cooldown_period:
            return

        threshold_memory_size = self._max_used_memory_ratio * self._max_memory_size
        buffer_memory_size = self._max_memory_size * (1 - self._max_used_memory_ratio) * self._reserve_memory_ratio
        overload_memory_threshold_size = threshold_memory_size + buffer_memory_size

        # Log a warning if current memory usage exceeds the critical overload threshold
        if current_memory_usage_size > overload_memory_threshold_size:
            memory_usage_percentage = round((current_memory_usage_size.bytes / self._max_memory_size.bytes) * 100)
            logger.warning(
                f'Memory is critically overloaded. Using {current_memory_usage_size} of '
                f'{self._max_memory_size} ({memory_usage_percentage}%). '
                'Consider increasing available memory.'
            )
            self._timestamp_of_last_memory_warning = snapshot_timestamp



from __future__ import annotations

from typing import Awaitable, Callable, Generic, TypeVar

from crawlee.basic_crawler.types import BasicCrawlingContext

TCrawlingContext = TypeVar('TCrawlingContext', bound=BasicCrawlingContext)
RequestHandler = Callable[[TCrawlingContext], Awaitable[None]]


class Router(Generic[TCrawlingContext]):
    """Dispatches requests to registered handlers based on their labels."""

    def __init__(self) -> None:
        self._default_handler: RequestHandler[TCrawlingContext] | None = None
        self._handlers_by_label = dict[str, RequestHandler[TCrawlingContext]]()

    def default_handler(self: Router, handler: RequestHandler[TCrawlingContext]) -> RequestHandler[TCrawlingContext]:
        """A decorator used to register a default handler.

        The default handler is invoked for requests that have either no label or a label for which we have no matching
        handler.
        """
        if self._default_handler is not None:
            raise RuntimeError('A default handler is already configured')

        self._default_handler = handler

        return handler

    def handler(self, label: str) -> Callable[[RequestHandler[TCrawlingContext]], None]:
        """A decorator used to register a label-based handler.

        The registered will be invoked only for requests with the exact same label.
        """
        if label in self._handlers_by_label:
            raise RuntimeError(f'A handler for label `{label}` is already registered')

        def wrapper(handler: Callable[[TCrawlingContext], Awaitable]) -> None:
            self._handlers_by_label[label] = handler

        return wrapper

    async def __call__(self, context: TCrawlingContext) -> None:
        """Invoke a request handler that matches the request label (or the default)."""
        if context.request.label is None or context.request.label not in self._handlers_by_label:
            if self._default_handler is None:
                raise RuntimeError(
                    f'No handler matches label `{context.request.label}` and no default handler is configured'
                )

            return await self._default_handler(context)

        handler = self._handlers_by_label[context.request.label]
        return await handler(context)



# ruff: noqa: TCH001 TCH003
from __future__ import annotations

import json
from dataclasses import asdict, dataclass
from datetime import datetime, timedelta
from typing import Annotated, Any

from pydantic import BaseModel, ConfigDict, Field
from typing_extensions import override

from crawlee._utils.models import timedelta_ms


@dataclass(frozen=True)
class FinalStatistics:
    """Statistics about a crawler run."""

    requests_finished: int
    requests_failed: int
    retry_histogram: list[int]
    request_avg_failed_duration: timedelta | None
    request_avg_finished_duration: timedelta | None
    requests_finished_per_minute: float
    requests_failed_per_minute: float
    request_total_duration: timedelta
    requests_total: int
    crawler_runtime: timedelta

    @override
    def __str__(self) -> str:
        return json.dumps(
            {k: v.total_seconds() if isinstance(v, timedelta) else v for k, v in asdict(self).items()},
            indent=2,
        )


class StatisticsState(BaseModel):
    """Statistic data about a crawler run."""

    model_config = ConfigDict(populate_by_name=True, ser_json_inf_nan='constants')

    requests_finished: Annotated[int, Field(alias='requestsFinished')] = 0
    requests_failed: Annotated[int, Field(alias='requestsFailed')] = 0
    requests_retries: Annotated[int, Field(alias='requestsRetries')] = 0
    requests_failed_per_minute: Annotated[float, Field(alias='requestsFailedPerMinute')] = 0
    requests_finished_per_minute: Annotated[float, Field(alias='requestsFinishedPerMinute')] = 0
    request_min_duration: Annotated[timedelta_ms | None, Field(alias='requestMinDurationMillis')] = None
    request_max_duration: Annotated[timedelta_ms | None, Field(alias='requestMaxDurationMillis')] = None
    request_total_failed_duration: Annotated[timedelta_ms, Field(alias='requestTotalFailedDurationMillis')] = (
        timedelta()
    )
    request_total_finished_duration: Annotated[timedelta_ms, Field(alias='requestTotalFinishedDurationMillis')] = (
        timedelta()
    )
    crawler_started_at: Annotated[datetime | None, Field(alias='crawlerStartedAt')] = None
    crawler_last_started_at: Annotated[datetime | None, Field(alias='crawlerLastStartTimestamp')] = None
    crawler_finished_at: Annotated[datetime | None, Field(alias='crawlerFinishedAt')] = None
    crawler_runtime: Annotated[timedelta_ms, Field(alias='crawlerRuntimeMillis')] = timedelta()
    errors: dict[str, Any] = Field(default_factory=dict)
    retry_errors: dict[str, Any] = Field(alias='retryErrors', default_factory=dict)
    requests_with_status_code: dict[str, int] = Field(alias='requestsWithStatusCode', default_factory=dict)
    stats_persisted_at: Annotated[datetime | None, Field(alias='statsPersistedAt')] = None


class StatisticsPersistedState(BaseModel):
    """Additional statistic data to be stored in the persisted state."""

    model_config = ConfigDict(populate_by_name=True)

    request_retry_histogram: Annotated[list[int], Field(alias='requestRetryHistogram')]
    stats_id: Annotated[int, Field(alias='statsId')]
    request_avg_failed_duration: Annotated[timedelta_ms | None, Field(alias='requestAvgFailedDurationMillis')]
    request_avg_finished_duration: Annotated[timedelta_ms | None, Field(alias='requestAvgFinishedDurationMillis')]
    request_total_duration: Annotated[timedelta_ms, Field(alias='requestTotalDurationMillis')]
    requests_total: Annotated[int, Field(alias='requestsTotal')]
    crawler_last_started_at: Annotated[datetime, Field(alias='crawlerLastStartTimestamp')]
    stats_persisted_at: Annotated[datetime, Field(alias='statsPersistedAt')]



from .basic_crawler import BasicCrawler, BasicCrawlerOptions
from .context_pipeline import ContextPipeline
from .router import Router
from .types import BasicCrawlingContext

__all__ = ['BasicCrawler', 'BasicCrawlerOptions', 'ContextPipeline', 'Router', 'BasicCrawlingContext']



from .autoscaled_pool import AutoscaledPool, ConcurrencySettings
from .snapshotter import Snapshotter
from .system_status import SystemStatus

__all__ = ['AutoscaledPool', 'ConcurrencySettings', 'Snapshotter', 'SystemStatus']



from __future__ import annotations

import asyncio
import contextlib
import io
import json
import mimetypes
import re
from enum import Enum
from typing import Any

import aioshutil
from aiofiles import ospath
from aiofiles.os import remove, rename


class ContentType(Enum):
    JSON = r'^application/json'
    TEXT = r'^text/'
    XML = r'^application/.*xml$'

    def matches(self, content_type: str) -> bool:
        """Check if the content type matches the enum's pattern."""
        return bool(re.search(self.value, content_type, re.IGNORECASE))


def is_content_type(content_type_enum: ContentType, content_type: str) -> bool:
    """Check if the provided content type string matches the specified ContentType."""
    return content_type_enum.matches(content_type)


async def force_remove(filename: str) -> None:
    """Removes a file, suppressing the FileNotFoundError if it does not exist.

    JS-like rm(filename, { force: true }).

    Args:
        filename: The path to the file to be removed.
    """
    with contextlib.suppress(FileNotFoundError):
        await remove(filename)


async def force_rename(src_dir: str, dst_dir: str) -> None:
    """Renames a directory, ensuring that the destination directory is removed if it exists.

    Args:
        src_dir: The source directory path.
        dst_dir: The destination directory path.
    """
    # Make sure source directory exists
    if await ospath.exists(src_dir):
        # Remove destination directory if it exists
        if await ospath.exists(dst_dir):
            await aioshutil.rmtree(dst_dir, ignore_errors=True)
        await rename(src_dir, dst_dir)


def determine_file_extension(content_type: str) -> str | None:
    """Determine the file extension for a given MIME content type.

    Args:
        content_type: The MIME content type string.

    Returns:
        A string representing the determined file extension without a leading dot,
            or None if no extension could be determined.
    """
    # e.g. mimetypes.guess_extension('application/json ') does not work...
    actual_content_type = content_type.split(';')[0].strip()

    # mimetypes.guess_extension returns 'xsl' in this case, because 'application/xxx' is "structured"
    # ('text/xml' would be "unstructured" and return 'xml') we have to explicitly override it here
    if actual_content_type == 'application/xml':
        return 'xml'

    # Determine the extension from the mime type
    ext = mimetypes.guess_extension(actual_content_type)

    # Remove the leading dot if extension successfully parsed
    return ext[1:] if ext is not None else ext


def is_file_or_bytes(value: Any) -> bool:
    """Determine if the input value is a file-like object or bytes.

    This function checks whether the provided value is an instance of bytes, bytearray, or io.IOBase (file-like).
    The method is simplified for common use cases and may not cover all edge cases.

    Args:
        value: The value to be checked.

    Returns:
        True if the value is either a file-like object or bytes, False otherwise.
    """
    return isinstance(value, (bytes, bytearray, io.IOBase))


async def json_dumps(obj: Any) -> str:
    """Serialize an object to a JSON-formatted string with specific settings.

    Args:
        obj: The object to serialize.

    Returns:
        A string containing the JSON representation of the input object.
    """
    return await asyncio.to_thread(json.dumps, obj, ensure_ascii=False, indent=2, default=str)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, cast

import httpx
from typing_extensions import override

from crawlee._utils.blocked import ROTATE_PROXY_ERRORS
from crawlee.basic_crawler.errors import ProxyError
from crawlee.http_clients import BaseHttpClient, HttpCrawlingResult, HttpResponse
from crawlee.sessions import Session

if TYPE_CHECKING:
    from collections.abc import Iterable

    from crawlee.models import Request
    from crawlee.proxy_configuration import ProxyInfo
    from crawlee.statistics import Statistics

__all__ = ['HttpxClient']


class HttpTransport(httpx.AsyncHTTPTransport):
    """A modified HTTP transport adapter that stores response cookies in a `Session` instead of the httpx client."""

    @override
    async def handle_async_request(self, request: httpx.Request) -> httpx.Response:
        response = await super().handle_async_request(request)
        response.request = request

        if session := cast(Session, request.extensions.get('crawlee_session')):
            response_cookies = httpx.Cookies()
            response_cookies.extract_cookies(response)
            session.cookies.update(response_cookies)

        if 'Set-Cookie' in response.headers:
            del response.headers['Set-Cookie']

        return response


def _is_proxy_error(error: httpx.TransportError) -> bool:
    if isinstance(error, httpx.ProxyError):
        return True

    if any(needle in str(error) for needle in ROTATE_PROXY_ERRORS):  # noqa: SIM103
        return True

    return False


class HttpxClient(BaseHttpClient):
    """A httpx-based HTTP client used for making HTTP calls in crawlers (`BasicCrawler` subclasses)."""

    def __init__(
        self,
        *,
        persist_cookies_per_session: bool = True,
        additional_http_error_status_codes: Iterable[int] = (),
        ignore_http_error_status_codes: Iterable[int] = (),
    ) -> None:
        super().__init__(
            persist_cookies_per_session=persist_cookies_per_session,
            additional_http_error_status_codes=additional_http_error_status_codes,
            ignore_http_error_status_codes=ignore_http_error_status_codes,
        )

        self._client_by_proxy_url = dict[Optional[str], httpx.AsyncClient]()

    def _get_client(self, proxy_url: str | None) -> httpx.AsyncClient:
        if proxy_url not in self._client_by_proxy_url:
            self._client_by_proxy_url[proxy_url] = httpx.AsyncClient(
                transport=HttpTransport(),
                proxy=proxy_url,
                timeout=httpx.Timeout(10),
            )

        return self._client_by_proxy_url[proxy_url]

    @override
    async def crawl(
        self,
        request: Request,
        session: Session | None,
        proxy_info: ProxyInfo | None,
        statistics: Statistics,
    ) -> HttpCrawlingResult:
        client = self._get_client(proxy_info.url if proxy_info else None)
        http_request = client.build_request(
            method=request.method,
            url=request.url,
            headers=request.headers,
            cookies=session.cookies if session else None,
            extensions={'crawlee_session': session if self._persist_cookies_per_session else None},
        )

        try:
            response = await client.send(http_request, follow_redirects=True)
        except httpx.TransportError as e:
            if _is_proxy_error(e):
                raise ProxyError from e

            raise

        statistics.register_status_code(response.status_code)

        exclude_error = response.status_code in self._ignore_http_error_status_codes
        include_error = response.status_code in self._additional_http_error_status_codes

        if include_error or (response.is_server_error and not exclude_error):
            if include_error:
                raise httpx.HTTPStatusError(
                    f'Status code {response.status_code} (user-configured to be an error) returned',
                    request=response.request,
                    response=response,
                )

            raise httpx.HTTPStatusError(
                f'Status code {response.status_code} returned', request=response.request, response=response
            )

        request.loaded_url = str(response.url)

        return HttpCrawlingResult(http_response=response)

    @override
    async def send_request(
        self,
        url: str,
        *,
        method: str,
        headers: httpx.Headers | dict[str, str],
        session: Session | None,
        proxy_info: ProxyInfo | None,
    ) -> HttpResponse:
        client = self._get_client(proxy_info.url if proxy_info else None)

        http_request = client.build_request(
            url=url,
            method=method,
            headers=headers,
            extensions={'crawlee_session': session if self._persist_cookies_per_session else None},
        )

        try:
            response = await client.send(http_request)
        except httpx.TransportError as e:
            if _is_proxy_error(e):
                raise ProxyError from e

            raise

        return response



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from crawlee.models import DatasetListPage, DatasetMetadata


class BaseDatasetCollectionClient(ABC):
    """Abstract base class for dataset collection clients.

    This collection client handles operations that involve multiple instances of a given resource type.
    """

    @abstractmethod
    async def get_or_create(
        self,
        *,
        id: str | None = None,
        name: str | None = None,
        schema: dict | None = None,
    ) -> DatasetMetadata:
        """Retrieve an existing dataset by its name or ID, or create a new one if it does not exist.

        Args:
            id: Optional ID of the dataset to retrieve or create. If provided, the method will attempt
                to find a dataset with the ID.

            name: Optional name of the dataset resource to retrieve or create. If provided, the method will
                attempt to find a dataset with this name.

            schema: Optional schema for the dataset resource to be created.

        Returns:
            Metadata object containing the information of the retrieved or created dataset.
        """

    @abstractmethod
    async def list(
        self,
        *,
        unnamed: bool = False,
        limit: int | None = None,
        offset: int | None = None,
        desc: bool = False,
    ) -> DatasetListPage:
        """List the available datasets.

        Args:
            unnamed: Whether to list only the unnamed datasets.
            limit: Maximum number of datasets to return.
            offset: Number of datasets to skip from the beginning of the list.
            desc: Whether to sort the datasets in descending order.

        Returns:
            The list of available datasets matching the specified filters.
        """



from __future__ import annotations

from typing import Union

from .base_dataset_client import BaseDatasetClient
from .base_dataset_collection_client import BaseDatasetCollectionClient
from .base_key_value_store_client import BaseKeyValueStoreClient
from .base_key_value_store_collection_client import BaseKeyValueStoreCollectionClient
from .base_request_queue_client import BaseRequestQueueClient
from .base_request_queue_collection_client import BaseRequestQueueCollectionClient

ResourceClient = Union[
    BaseDatasetClient,
    BaseKeyValueStoreClient,
    BaseRequestQueueClient,
]

ResourceCollectionClient = Union[
    BaseDatasetCollectionClient,
    BaseKeyValueStoreCollectionClient,
    BaseRequestQueueCollectionClient,
]



from __future__ import annotations

from enum import Enum


class EnqueueStrategy(str, Enum):
    """Strategy for deciding which links should be followed and which ones should be ignored."""

    ALL = 'all'
    SAME_DOMAIN = 'same-domain'
    SAME_HOSTNAME = 'same-hostname'
    SAME_ORIGIN = 'same-origin'



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, AsyncContextManager, AsyncIterator

if TYPE_CHECKING:
    from httpx import Response

    from crawlee.models import DatasetItemsListPage, DatasetMetadata
    from crawlee.types import JSONSerializable


class BaseDatasetClient(ABC):
    """Abstract base class for dataset resource clients.

    These clients are specific to the type of resource they manage and operate under a designated storage
    client, like a memory storage client.
    """

    _LIST_ITEMS_LIMIT = 999_999_999_999
    """This is what API returns in the x-apify-pagination-limit header when no limit query parameter is used."""

    @abstractmethod
    async def get(self) -> DatasetMetadata | None:
        """Get metadata about the dataset being managed by this client.

        Returns:
            An object containing the dataset's details, or None if the dataset does not exist.
        """

    @abstractmethod
    async def update(
        self,
        *,
        name: str | None = None,
    ) -> DatasetMetadata:
        """Update the dataset metadata.

        Args:
            name: New new name for the dataset.

        Returns:
            An object reflecting the updated dataset metadata.
        """

    @abstractmethod
    async def delete(self) -> None:
        """Permanently delete the dataset managed by this client."""

    @abstractmethod
    async def list_items(
        self,
        *,
        offset: int | None = 0,
        limit: int | None = _LIST_ITEMS_LIMIT,
        clean: bool = False,
        desc: bool = False,
        fields: list[str] | None = None,
        omit: list[str] | None = None,
        unwind: str | None = None,
        skip_empty: bool = False,
        skip_hidden: bool = False,
        flatten: list[str] | None = None,
        view: str | None = None,
    ) -> DatasetItemsListPage:
        """Retrieves a paginated list of items from a dataset based on various filtering parameters.

        This method provides the flexibility to filter, sort, and modify the appearance of dataset items
        when listed. Each parameter modifies the result set according to its purpose. The method also
        supports pagination through 'offset' and 'limit' parameters.

        Args:
            offset: The number of initial items to skip.
            limit: The maximum number of items to return.
            clean: If True, removes empty items and hidden fields, equivalent to 'skip_hidden' and 'skip_empty'.
            desc: If True, items are returned in descending order, i.e., newest first.
            fields: Specifies a subset of fields to include in each item.
            omit: Specifies a subset of fields to exclude from each item.
            unwind: Specifies a field that should be unwound. If it's an array, each element becomes a separate record.
            skip_empty: If True, omits items that are empty after other filters have been applied.
            skip_hidden: If True, omits fields starting with the '#' character.
            flatten: A list of fields to flatten in each item.
            view: The specific view of the dataset to use when retrieving items.

        Returns:
            An object with filtered, sorted, and paginated dataset items plus pagination details.
        """

    @abstractmethod
    async def iterate_items(
        self,
        *,
        offset: int = 0,
        limit: int | None = None,
        clean: bool = False,
        desc: bool = False,
        fields: list[str] | None = None,
        omit: list[str] | None = None,
        unwind: str | None = None,
        skip_empty: bool = False,
        skip_hidden: bool = False,
    ) -> AsyncIterator[dict]:
        """Iterates over items in the dataset according to specified filters and sorting.

        This method allows for asynchronously iterating through dataset items while applying various filters such as
        skipping empty items, hiding specific fields, and sorting. It supports pagination via `offset` and `limit`
        parameters, and can modify the appearance of dataset items using `fields`, `omit`, `unwind`, `skip_empty`, and
        `skip_hidden` parameters.

        Args:
            offset: The number of initial items to skip.
            limit: The maximum number of items to iterate over. Defaults to no limit.
            clean: If True, removes empty items and hidden fields, equivalent to 'skip_hidden' and 'skip_empty'.
            desc: If set to True, items are returned in descending order, i.e., newest first.
            fields: Specifies a subset of fields to include in each item.
            omit: Specifies a subset of fields to exclude from each item.
            unwind: Specifies a field that should be unwound into separate items.
            skip_empty: If set to True, omits items that are empty after other filters have been applied.
            skip_hidden: If set to True, omits fields starting with the '#' character from the output.

        Yields:
            An asynchronous iterator of dictionary objects, each representing a dataset item after applying
            the specified filters and transformations.
        """

    @abstractmethod
    async def get_items_as_bytes(
        self,
        *,
        item_format: str = 'json',
        offset: int | None = None,
        limit: int | None = None,
        desc: bool = False,
        clean: bool = False,
        bom: bool = False,
        delimiter: str | None = None,
        fields: list[str] | None = None,
        omit: list[str] | None = None,
        unwind: str | None = None,
        skip_empty: bool = False,
        skip_header_row: bool = False,
        skip_hidden: bool = False,
        xml_root: str | None = None,
        xml_row: str | None = None,
        flatten: list[str] | None = None,
    ) -> bytes:
        """Retrieves dataset items as bytes.

        Args:
            item_format: Output format (e.g., 'json', 'csv'); default is 'json'.
            offset: Number of items to skip; default is 0.
            limit: Max number of items to return; no default limit.
            desc: If True, results are returned in descending order.
            clean: If True, filters out empty items and hidden fields.
            bom: Include or exclude UTF-8 BOM; default behavior varies by format.
            delimiter: Delimiter character for CSV; default is ','.
            fields: List of fields to include in the results.
            omit: List of fields to omit from the results.
            unwind: Unwinds a field into separate records.
            skip_empty: If True, skips empty items in the output.
            skip_header_row: If True, skips the header row in CSV.
            skip_hidden: If True, skips hidden fields in the output.
            xml_root: Root element name for XML output; default is 'items'.
            xml_row: Element name for each item in XML output; default is 'item'.
            flatten: List of fields to flatten.

        Returns:
            The dataset items as raw bytes.
        """

    @abstractmethod
    async def stream_items(
        self,
        *,
        item_format: str = 'json',
        offset: int | None = None,
        limit: int | None = None,
        desc: bool = False,
        clean: bool = False,
        bom: bool = False,
        delimiter: str | None = None,
        fields: list[str] | None = None,
        omit: list[str] | None = None,
        unwind: str | None = None,
        skip_empty: bool = False,
        skip_header_row: bool = False,
        skip_hidden: bool = False,
        xml_root: str | None = None,
        xml_row: str | None = None,
    ) -> AsyncContextManager[Response | None]:
        """Retrieves dataset items as a streaming response.

        Args:
            item_format: Output format, options include json, jsonl, csv, html, xlsx, xml, rss; default is json.
            offset: Number of items to skip at the start; default is 0.
            limit: Maximum number of items to return; no default limit.
            desc: If True, reverses the order of results.
            clean: If True, filters out empty items and hidden fields.
            bom: Include or exclude UTF-8 BOM; varies by format.
            delimiter: Delimiter for CSV files; default is ','.
            fields: List of fields to include in the output.
            omit: List of fields to omit from the output.
            unwind: Unwinds a field into separate records.
            skip_empty: If True, empty items are omitted.
            skip_header_row: If True, skips the header row in CSV.
            skip_hidden: If True, hides fields starting with the # character.
            xml_root: Custom root element name for XML output; default is 'items'.
            xml_row: Custom element name for each item in XML; default is 'item'.

        Yields:
            The dataset items in a streaming response.
        """

    @abstractmethod
    async def push_items(self, items: JSONSerializable) -> None:
        """Push items to the dataset.

        Args:
            items: The items which to push in the dataset. They must be JSON serializable.
        """



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Iterable, Protocol

if TYPE_CHECKING:
    from httpx import Headers  # Type from `httpx` is used here because it is lightweight and convenient

    from crawlee.models import Request
    from crawlee.proxy_configuration import ProxyInfo
    from crawlee.sessions import Session
    from crawlee.statistics import Statistics


class HttpResponse(Protocol):
    """Response to an HTTP call."""

    def read(self) -> bytes:
        """Read the content of the response body."""

    @property
    def status_code(self) -> int:
        """HTTP status code of the response (https://developer.mozilla.org/en-US/docs/Web/HTTP/Status)."""

    @property
    def headers(self) -> Headers:
        """Response headers."""


@dataclass(frozen=True)
class HttpCrawlingResult:
    """Result of a HTTP-only crawl."""

    http_response: HttpResponse


class BaseHttpClient(ABC):
    """An HTTP client used for making HTTP calls in crawlers (`BasicCrawler` subclasses)."""

    def __init__(
        self,
        *,
        persist_cookies_per_session: bool = True,
        additional_http_error_status_codes: Iterable[int] = (),
        ignore_http_error_status_codes: Iterable[int] = (),
    ) -> None:
        self._persist_cookies_per_session = persist_cookies_per_session
        self._additional_http_error_status_codes = set(additional_http_error_status_codes)
        self._ignore_http_error_status_codes = set(ignore_http_error_status_codes)

    @abstractmethod
    async def crawl(
        self,
        request: Request,
        session: Session | None,
        proxy_info: ProxyInfo | None,
        statistics: Statistics,
    ) -> HttpCrawlingResult:
        """Perform a crawl of an URL."""

    @abstractmethod
    async def send_request(
        self,
        url: str,
        *,
        method: str,
        headers: Headers | dict[str, str],
        session: Session | None,
        proxy_info: ProxyInfo | None,
    ) -> HttpResponse:
        """Perform an HTTP request."""



from __future__ import annotations

from typing import TYPE_CHECKING

from crawlee.memory_storage_client import MemoryStorageClient

if TYPE_CHECKING:
    from crawlee.base_storage_client import BaseStorageClient


class StorageClientManager:
    """A class for managing storage clients."""

    _local_client: BaseStorageClient = MemoryStorageClient()
    _cloud_client: BaseStorageClient | None = None

    @classmethod
    def get_storage_client(cls, *, in_cloud: bool = False) -> BaseStorageClient:
        """Get the storage client instance for the current environment.

        Args:
            in_cloud: Whether the code is running in the cloud environment.

        Returns:
            The current storage client instance.
        """
        if in_cloud:
            if cls._cloud_client is None:
                raise RuntimeError('Running in cloud environment, but cloud client was not provided.')
            return cls._cloud_client

        return cls._local_client

    @classmethod
    def set_cloud_client(cls, cloud_client: BaseStorageClient) -> None:
        """Set the cloud storage client instance.

        Args:
            cloud_client: The cloud storage client instance.
        """
        cls._cloud_client = cloud_client

    @classmethod
    def set_local_client(cls, local_client: BaseStorageClient) -> None:
        """Set the local storage client instance.

        Args:
            local_client: The local storage client instance.
        """
        cls._local_client = local_client



from .dataset import Dataset
from .key_value_store import KeyValueStore
from .request_list import RequestList
from .request_queue import RequestQueue

__all__ = ['Dataset', 'KeyValueStore', 'RequestList', 'RequestQueue']



# Inspiration: https://github.com/apify/crawlee/blob/v3.10.1/packages/browser-pool/src/abstract-classes/browser-controller.ts

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import Mapping
    from datetime import datetime, timedelta

    from playwright.async_api import Page

    from crawlee.proxy_configuration import ProxyInfo


class BaseBrowserController(ABC):
    """An abstract class for managing browser instance and their pages."""

    AUTOMATION_LIBRARY: str | None = None
    """The name of the automation library that the controller is using."""

    @property
    @abstractmethod
    def pages(self) -> list[Page]:
        """Return the list of opened pages."""

    @property
    @abstractmethod
    def pages_count(self) -> int:
        """Returns the number of currently open pages."""

    @property
    @abstractmethod
    def last_page_opened_at(self) -> datetime:
        """Return the time when the last page was opened."""

    @property
    @abstractmethod
    def idle_time(self) -> timedelta:
        """Return the idle time of the browser controller."""

    @property
    @abstractmethod
    def has_free_capacity(self) -> bool:
        """Return if the browser has free capacity to open a new page."""

    @property
    @abstractmethod
    def is_browser_connected(self) -> bool:
        """Return if the browser is closed."""

    @abstractmethod
    async def new_page(
        self,
        page_options: Mapping[str, Any] | None = None,
        proxy_info: ProxyInfo | None = None,
    ) -> Page:
        """Create a new page with the given context options.

        Args:
            page_options: Options to configure the new page.
            proxy_info: The proxy configuration to use for the new page.

        Returns:
            Page: The newly created page.

        Raises:
            ValueError: If the browser has reached the maximum number of open pages.
        """

    @abstractmethod
    async def close(self, *, force: bool = False) -> None:
        """Close the browser.

        Args:
            force: Whether to force close all open pages before closing the browser.

        Raises:
            ValueError: If there are still open pages when trying to close the browser.
        """



try:
    from .browser_pool import BrowserPool
    from .playwright_browser_controller import PlaywrightBrowserController
    from .playwright_browser_plugin import PlaywrightBrowserPlugin
except ImportError as exc:
    raise ImportError(
        'To import anything from this subpacakge, you need to install the "playwright" extra. '
        'For example, if you use pip, run "pip install crawlee[playwright]".',
    ) from exc

__all__ = ['BrowserPool', 'PlaywrightBrowserController', 'PlaywrightBrowserPlugin']



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from crawlee.models import RequestQueueListPage, RequestQueueMetadata


class BaseRequestQueueCollectionClient(ABC):
    """Abstract base class for request queue collection clients.

    This collection client handles operations that involve multiple instances of a given resource type.
    """

    @abstractmethod
    async def get_or_create(
        self,
        *,
        id: str | None = None,
        name: str | None = None,
        schema: dict | None = None,
    ) -> RequestQueueMetadata:
        """Retrieve an existing request queue by its name or ID, or create a new one if it does not exist.

        Args:
            id: Optional ID of the request queue to retrieve or create. If provided, the method will attempt
                to find a request queue with the ID.

            name: Optional name of the request queue resource to retrieve or create. If provided, the method will
                attempt to find a request queue with this name.

            schema: Optional schema for the request queue resource to be created.

        Returns:
            Metadata object containing the information of the retrieved or created request queue.
        """

    @abstractmethod
    async def list(
        self,
        *,
        unnamed: bool = False,
        limit: int | None = None,
        offset: int | None = None,
        desc: bool = False,
    ) -> RequestQueueListPage:
        """List the available request queues.

        Args:
            unnamed: Whether to list only the unnamed request queues.
            limit: Maximum number of request queues to return.
            offset: Number of request queues to skip from the beginning of the list.
            desc: Whether to sort the request queues in descending order.

        Returns:
            The list of available request queues matching the specified filters.
        """



# Inspiration: https://github.com/apify/crawlee/blob/v3.10.1/packages/browser-pool/src/playwright/playwright-controller.ts

from __future__ import annotations

from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Any

from playwright.async_api import Page
from typing_extensions import override

from crawlee.browsers.base_browser_controller import BaseBrowserController

if TYPE_CHECKING:
    from collections.abc import Mapping

    from playwright.async_api import Browser

    from crawlee.proxy_configuration import ProxyInfo


class PlaywrightBrowserController(BaseBrowserController):
    """Controller for managing Playwright browser instances and their pages.

    This class provides methods to manage pages within a browser instance, ensuring that the number
    of open pages does not exceed the specified limit and tracking the state of the pages.
    """

    AUTOMATION_LIBRARY = 'playwright'

    def __init__(self, browser: Browser, *, max_open_pages_per_browser: int = 20) -> None:
        """Create a new instance.

        Args:
            browser: The browser instance to control.
            max_open_pages_per_browser: The maximum number of pages that can be open at the same time.
        """
        self._browser = browser
        self._max_open_pages_per_browser = max_open_pages_per_browser

        self._pages = list[Page]()
        self._last_page_opened_at = datetime.now(timezone.utc)

    @property
    @override
    def pages(self) -> list[Page]:
        return self._pages

    @property
    @override
    def pages_count(self) -> int:
        return len(self._pages)

    @property
    @override
    def last_page_opened_at(self) -> datetime:
        return self._last_page_opened_at

    @property
    @override
    def idle_time(self) -> timedelta:
        return datetime.now(timezone.utc) - self._last_page_opened_at

    @property
    @override
    def has_free_capacity(self) -> bool:
        return self.pages_count < self._max_open_pages_per_browser

    @property
    @override
    def is_browser_connected(self) -> bool:
        return self._browser.is_connected()

    @override
    async def new_page(
        self,
        page_options: Mapping[str, Any] | None = None,
        proxy_info: ProxyInfo | None = None,
    ) -> Page:
        page_options = dict(page_options) if page_options else {}

        # If "proxy_info" is provided and no proxy is already set in "page_options", configure the proxy.
        if proxy_info and 'proxy' not in page_options:
            page_options['proxy'] = {
                'server': f'{proxy_info.scheme}://{proxy_info.hostname}:{proxy_info.port}',
                'username': proxy_info.username,
                'password': proxy_info.password,
            }

        if not self.has_free_capacity:
            raise ValueError('Cannot open more pages in this browser.')

        page = await self._browser.new_page(**page_options)

        # Handle page close event
        page.on(event='close', f=self._on_page_close)

        # Update internal state
        self._pages.append(page)
        self._last_page_opened_at = datetime.now(timezone.utc)

        return page

    @override
    async def close(self, *, force: bool = False) -> None:
        if force:
            for page in self._pages:
                await page.close()

        if self.pages_count > 0:
            raise ValueError('Cannot close the browser while there are open pages.')

        await self._browser.close()

    def _on_page_close(self, page: Page) -> None:
        """Handle actions after a page is closed."""
        self._pages.remove(page)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from crawlee.configuration import Configuration


class BaseStorage(ABC):
    """Base class for storages."""

    LABEL = 'Unknown'
    """Human readable label of the storage."""

    @property
    @abstractmethod
    def id(self) -> str:
        """Get the storage ID."""

    @property
    @abstractmethod
    def name(self) -> str | None:
        """Get the storage name."""

    @classmethod
    @abstractmethod
    async def open(
        cls,
        *,
        id: str | None = None,
        name: str | None = None,
        configuration: Configuration | None = None,
    ) -> BaseStorage:
        """Open a storage, either restore existing or create a new one.

        Args:
            id: The storage ID.
            name: The storage name.
            configuration: The configuration to use.
        """

    @abstractmethod
    async def drop(self) -> None:
        """Drop the storage. Remove it from underlying storage and delete from cache."""



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import timedelta
from typing import TYPE_CHECKING

from crawlee.models import BaseRequestData, Request

if TYPE_CHECKING:
    from collections.abc import Sequence

    from crawlee.models import ProcessedRequest


class RequestProvider(ABC):
    """Provides access to a queue of crawling requests."""

    @property
    @abstractmethod
    def name(self) -> str | None:
        """ID or name of the request queue."""

    @abstractmethod
    async def get_total_count(self) -> int:
        """Returns an offline approximation of the total number of requests in the queue (i.e. pending + handled)."""

    @abstractmethod
    async def is_empty(self) -> bool:
        """Returns True if there are no more requests in the queue (there might still be unfinished requests)."""

    @abstractmethod
    async def is_finished(self) -> bool:
        """Returns True if all requests have been handled."""

    @abstractmethod
    async def drop(self) -> None:
        """Removes the queue either from the Apify Cloud storage or from the local database."""

    @abstractmethod
    async def fetch_next_request(self) -> Request | None:
        """Returns a next request in the queue to be processed, or `null` if there are no more pending requests."""

    @abstractmethod
    async def reclaim_request(self, request: Request, *, forefront: bool = False) -> ProcessedRequest | None:
        """Reclaims a failed request back to the queue, so that it can be returned for processing later again.

        It is possible to modify the request data by supplying an updated request as a parameter.
        """

    @abstractmethod
    async def mark_request_as_handled(self, request: Request) -> ProcessedRequest | None:
        """Marks a request as handled after a successful processing (or after giving up retrying)."""

    @abstractmethod
    async def get_handled_count(self) -> int:
        """Returns the number of handled requests."""

    @abstractmethod
    async def add_requests_batched(
        self,
        requests: Sequence[str | BaseRequestData | Request],
        *,
        batch_size: int = 1000,
        wait_time_between_batches: timedelta = timedelta(seconds=1),
        wait_for_all_requests_to_be_added: bool = False,
        wait_for_all_requests_to_be_added_timeout: timedelta | None = None,
    ) -> None:
        """Add requests to the underlying resource client in batches.

        Args:
            requests: Requests to add to the queue.
            batch_size: The number of requests to add in one batch.
            wait_time_between_batches: Time to wait between adding batches.
            wait_for_all_requests_to_be_added: If True, wait for all requests to be added before returning.
            wait_for_all_requests_to_be_added_timeout: Timeout for waiting for all requests to be added.
        """

    def _transform_request(self, request: str | BaseRequestData | Request) -> Request:
        """Transforms a request-like object into a Request object."""
        if isinstance(request, Request):
            return request

        if isinstance(request, str):
            return Request.from_url(request)

        if isinstance(request, BaseRequestData):
            return Request.from_base_request_data(request)

        raise ValueError(f'Invalid request type: {type(request)}')

    def _transform_requests(self, requests: Sequence[str | BaseRequestData | Request]) -> list[Request]:
        """Transforms a list of request-like objects into a list of Request objects."""
        processed_requests: list[Request] = []

        for request in requests:
            processed_request = self._transform_request(request)
            processed_requests.append(processed_request)

        return processed_requests



from __future__ import annotations

from typing import TYPE_CHECKING

from typing_extensions import override

from crawlee.base_storage_client import BaseKeyValueStoreCollectionClient
from crawlee.memory_storage_client._creation_management import get_or_create_inner
from crawlee.memory_storage_client.key_value_store_client import KeyValueStoreClient
from crawlee.models import KeyValueStoreListPage, KeyValueStoreMetadata

if TYPE_CHECKING:
    from crawlee.memory_storage_client.memory_storage_client import MemoryStorageClient


class KeyValueStoreCollectionClient(BaseKeyValueStoreCollectionClient):
    """Subclient for manipulating key-value stores."""

    def __init__(self, *, memory_storage_client: MemoryStorageClient) -> None:
        self._memory_storage_client = memory_storage_client

    @property
    def _storage_client_cache(self) -> list[KeyValueStoreClient]:
        return self._memory_storage_client.key_value_stores_handled

    @override
    async def get_or_create(
        self,
        *,
        name: str | None = None,
        schema: dict | None = None,
        id: str | None = None,
    ) -> KeyValueStoreMetadata:
        resource_client = await get_or_create_inner(
            memory_storage_client=self._memory_storage_client,
            storage_client_cache=self._storage_client_cache,
            resource_client_class=KeyValueStoreClient,
            name=name,
            id=id,
        )
        return resource_client.resource_info

    @override
    async def list(
        self,
        *,
        unnamed: bool = False,
        limit: int | None = None,
        offset: int | None = None,
        desc: bool = False,
    ) -> KeyValueStoreListPage:
        items = [storage.resource_info for storage in self._storage_client_cache]

        return KeyValueStoreListPage(
            total=len(items),
            count=len(items),
            offset=0,
            limit=len(items),
            desc=False,
            items=sorted(items, key=lambda item: item.created_at),
        )



from __future__ import annotations

from dataclasses import dataclass

from crawlee.basic_crawler.types import BasicCrawlingContext
from crawlee.http_clients import HttpCrawlingResult


@dataclass(frozen=True)
class HttpCrawlingContext(BasicCrawlingContext, HttpCrawlingResult):
    """HTTP crawling context."""



# ruff: noqa: TCH001 TCH002
from __future__ import annotations

from collections.abc import Callable, Coroutine
from enum import Enum
from typing import Annotated, Any, Union

from pydantic import BaseModel, ConfigDict, Field

from crawlee._utils.system import CpuInfo, MemoryInfo


class Event(str, Enum):
    """Enum of all possible events that can be emitted."""

    # Core events
    PERSIST_STATE = 'persistState'
    SYSTEM_INFO = 'systemInfo'
    MIGRATING = 'migrating'
    ABORTING = 'aborting'
    EXIT = 'exit'

    # Session pool events
    SESSION_RETIRED = 'sessionRetired'

    # Browser pool events
    BROWSER_LAUNCHED = 'browserLaunched'
    BROWSER_RETIRED = 'browserRetired'
    BROWSER_CLOSED = 'browserClosed'
    PAGE_CREATED = 'pageCreated'
    PAGE_CLOSED = 'pageClosed'


class EventPersistStateData(BaseModel):
    """Data for the persist state event."""

    model_config = ConfigDict(populate_by_name=True)

    is_migrating: Annotated[bool, Field(alias='isMigrating')]


class EventSystemInfoData(BaseModel):
    """Data for the system info event."""

    model_config = ConfigDict(populate_by_name=True)

    cpu_info: Annotated[CpuInfo, Field(alias='cpuInfo')]
    memory_info: Annotated[MemoryInfo, Field(alias='memoryInfo')]


class EventMigratingData(BaseModel):
    """Data for the migrating event."""

    model_config = ConfigDict(populate_by_name=True)


class EventAbortingData(BaseModel):
    """Data for the aborting event."""

    model_config = ConfigDict(populate_by_name=True)


class EventExitData(BaseModel):
    """Data for the exit event."""

    model_config = ConfigDict(populate_by_name=True)


EventData = Union[EventPersistStateData, EventSystemInfoData, EventMigratingData, EventAbortingData, EventExitData]
SyncListener = Callable[..., None]
AsyncListener = Callable[..., Coroutine[Any, Any, None]]
Listener = Union[SyncListener, AsyncListener]
WrappedListener = Callable[..., Coroutine[Any, Any, None]]



# ruff: noqa: TCH003
from __future__ import annotations

import logging
import re
from collections.abc import Coroutine, Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Protocol

from typing_extensions import NotRequired, TypedDict, Unpack

if TYPE_CHECKING:
    from crawlee import Glob
    from crawlee.enqueue_strategy import EnqueueStrategy
    from crawlee.http_clients.base_http_client import HttpResponse
    from crawlee.models import BaseRequestData, DatasetItemsListPage, Request
    from crawlee.proxy_configuration import ProxyInfo
    from crawlee.sessions.session import Session
    from crawlee.storages.dataset import ExportToKwargs, GetDataKwargs, PushDataKwargs
    from crawlee.types import JSONSerializable


class AddRequestsKwargs(TypedDict):
    """Keyword arguments for crawler's `add_requests` method."""

    limit: NotRequired[int]
    base_url: NotRequired[str]
    strategy: NotRequired[EnqueueStrategy]
    include: NotRequired[list[re.Pattern | Glob]]
    exclude: NotRequired[list[re.Pattern | Glob]]


class AddRequestsFunction(Protocol):
    """Type of a function for adding URLs to the request queue with optional filtering.

    This helper method simplifies the process of adding requests to the request provider. It opens the specified
    request provider and adds the requests to it.
    """

    def __call__(  # noqa: D102
        self,
        requests: Sequence[str | BaseRequestData | Request],
        **kwargs: Unpack[AddRequestsKwargs],
    ) -> Coroutine[None, None, None]: ...


class GetDataFunction(Protocol):
    """Type of a function for getting data from the dataset.

    This helper method simplifies the process of retrieving data from a dataset. It opens the specified
    dataset and then retrieves the data based on the provided parameters.
    """

    def __call__(  # noqa: D102
        self,
        dataset_id: str | None = None,
        dataset_name: str | None = None,
        **kwargs: Unpack[GetDataKwargs],
    ) -> Coroutine[None, None, DatasetItemsListPage]: ...


class PushDataFunction(Protocol):
    """Type of a function for pushing data to the dataset.

    This helper method simplifies the process of pushing data to a dataset. It opens the specified
    dataset and then pushes the provided data to it.
    """

    def __call__(  # noqa: D102
        self,
        data: JSONSerializable,
        dataset_id: str | None = None,
        dataset_name: str | None = None,
        **kwargs: Unpack[PushDataKwargs],
    ) -> Coroutine[None, None, None]: ...


class ExportToFunction(Protocol):
    """Type of a function for exporting data from a dataset.

    This helper method simplifies the process of exporting data from a dataset. It opens the specified
    dataset and then exports its content to the key-value store.
    """

    def __call__(  # noqa: D102
        self,
        dataset_id: str | None = None,
        dataset_name: str | None = None,
        **kwargs: Unpack[ExportToKwargs],
    ) -> Coroutine[None, None, None]: ...


class EnqueueLinksFunction(Protocol):
    """Type of a function for enqueueing links based on a selector.

    Args:
        selector: CSS selector used to find the elements containing the links.
        label: Label for the newly created `Request` objects, used for request routing.
        user_data: User data to be provided to the newly created `Request` objects.
        **kwargs: Additional arguments for the `add_requests` method.
    """

    def __call__(  # noqa: D102
        self,
        *,
        selector: str = 'a',
        label: str | None = None,
        user_data: dict[str, Any] | None = None,
        **kwargs: Unpack[AddRequestsKwargs],
    ) -> Coroutine[None, None, None]: ...


class SendRequestFunction(Protocol):
    """Type of a function for performing an HTTP request."""

    def __call__(  # noqa: D102
        self,
        url: str,
        *,
        method: str = 'get',
        headers: dict[str, str] | None = None,
    ) -> Coroutine[None, None, HttpResponse]: ...


@dataclass(frozen=True)
class BasicCrawlingContext:
    """Basic crawling context intended to be extended by crawlers."""

    request: Request
    session: Session | None
    proxy_info: ProxyInfo | None
    send_request: SendRequestFunction
    add_requests: AddRequestsFunction
    push_data: PushDataFunction
    log: logging.Logger


class AddRequestsFunctionCall(AddRequestsKwargs):
    """Record of a call to `add_requests`."""

    requests: Sequence[str | BaseRequestData | Request]


@dataclass()
class RequestHandlerRunResult:
    """Record of calls to storage-related context helpers."""

    add_requests_calls: list[AddRequestsFunctionCall] = field(default_factory=list)

    async def add_requests(
        self,
        requests: Sequence[str | BaseRequestData],
        **kwargs: Unpack[AddRequestsKwargs],
    ) -> None:
        """Track a call to the `add_requests` context helper."""
        self.add_requests_calls.append(AddRequestsFunctionCall(requests=requests, **kwargs))



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, AsyncGenerator, Iterable

from typing_extensions import Unpack

from crawlee.basic_crawler import BasicCrawler, BasicCrawlerOptions, ContextPipeline
from crawlee.basic_crawler.errors import SessionError
from crawlee.http_clients import HttpxClient
from crawlee.http_crawler.types import HttpCrawlingContext

if TYPE_CHECKING:
    from crawlee.basic_crawler.types import BasicCrawlingContext


class HttpCrawler(BasicCrawler[HttpCrawlingContext]):
    """A crawler that fetches the request URL using `httpx`."""

    def __init__(
        self,
        *,
        additional_http_error_status_codes: Iterable[int] = (),
        ignore_http_error_status_codes: Iterable[int] = (),
        **kwargs: Unpack[BasicCrawlerOptions[HttpCrawlingContext]],
    ) -> None:
        """Initialize the HttpCrawler.

        Args:
            additional_http_error_status_codes: HTTP status codes that should be considered errors (and trigger a retry)

            ignore_http_error_status_codes: HTTP status codes that are normally considered errors but we want to treat
                them as successful

            kwargs: Arguments to be forwarded to the underlying BasicCrawler
        """
        kwargs['_context_pipeline'] = (
            ContextPipeline().compose(self._make_http_request).compose(self._handle_blocked_request)
        )

        kwargs.setdefault(
            'http_client',
            HttpxClient(
                additional_http_error_status_codes=additional_http_error_status_codes,
                ignore_http_error_status_codes=ignore_http_error_status_codes,
            ),
        )

        kwargs.setdefault('_logger', logging.getLogger(__name__))

        super().__init__(**kwargs)

    async def _make_http_request(self, context: BasicCrawlingContext) -> AsyncGenerator[HttpCrawlingContext, None]:
        result = await self._http_client.crawl(
            context.request,
            context.session,
            context.proxy_info,
            self._statistics,
        )

        yield HttpCrawlingContext(
            request=context.request,
            session=context.session,
            proxy_info=context.proxy_info,
            add_requests=context.add_requests,
            send_request=context.send_request,
            push_data=context.push_data,
            log=context.log,
            http_response=result.http_response,
        )

    async def _handle_blocked_request(self, context: HttpCrawlingContext) -> AsyncGenerator[HttpCrawlingContext, None]:
        if self._retry_on_blocked:
            status_code = context.http_response.status_code

            if context.session and context.session.is_blocked_status_code(status_code=status_code):
                raise SessionError(f'Assuming the session is blocked based on HTTP status code {status_code}')

        yield context



# Inspiration: https://github.com/apify/crawlee/blob/v3.10.1/packages/browser-pool/src/playwright/playwright-plugin.ts

from __future__ import annotations

from logging import getLogger
from typing import TYPE_CHECKING, Any, Literal

from playwright.async_api import Playwright, async_playwright
from typing_extensions import override

from crawlee.browsers.base_browser_plugin import BaseBrowserPlugin
from crawlee.browsers.playwright_browser_controller import PlaywrightBrowserController

if TYPE_CHECKING:
    from collections.abc import Mapping
    from types import TracebackType

logger = getLogger(__name__)


class PlaywrightBrowserPlugin(BaseBrowserPlugin):
    """A plugin for managing Playwright automation library.

    It should work as a factory for creating new browser instances.
    """

    AUTOMATION_LIBRARY = 'playwright'

    def __init__(
        self,
        *,
        browser_type: Literal['chromium', 'firefox', 'webkit'] = 'chromium',
        browser_options: Mapping[str, Any] | None = None,
        page_options: Mapping[str, Any] | None = None,
        max_open_pages_per_browser: int = 20,
    ) -> None:
        """Create a new instance.

        Args:
            browser_type: The type of the browser to launch.
            browser_options: Options to configure the browser instance.
            page_options: Options to configure a new page instance.
            max_open_pages_per_browser: The maximum number of pages that can be opened in a single browser instance.
                Once reached, a new browser instance will be launched to handle the excess.
        """
        self._browser_type = browser_type
        self._browser_options = browser_options or {}
        self._page_options = page_options or {}
        self._max_open_pages_per_browser = max_open_pages_per_browser

        self._playwright_context_manager = async_playwright()
        self._playwright: Playwright | None = None

    @property
    @override
    def browser_type(self) -> Literal['chromium', 'firefox', 'webkit']:
        return self._browser_type

    @property
    @override
    def browser_options(self) -> Mapping[str, Any]:
        return self._browser_options

    @property
    @override
    def page_options(self) -> Mapping[str, Any]:
        return self._page_options

    @property
    @override
    def max_open_pages_per_browser(self) -> int:
        return self._max_open_pages_per_browser

    @override
    async def __aenter__(self) -> PlaywrightBrowserPlugin:
        logger.debug('Initializing Playwright browser plugin.')
        self._playwright = await self._playwright_context_manager.__aenter__()
        return self

    @override
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        exc_traceback: TracebackType | None,
    ) -> None:
        logger.debug('Closing Playwright browser plugin.')
        await self._playwright_context_manager.__aexit__(exc_type, exc_value, exc_traceback)

    @override
    async def new_browser(self) -> PlaywrightBrowserController:
        if not self._playwright:
            raise RuntimeError('Playwright browser plugin is not initialized.')

        if self._browser_type == 'chromium':
            browser = await self._playwright.chromium.launch(**self._browser_options)
        elif self._browser_type == 'firefox':
            browser = await self._playwright.firefox.launch(**self._browser_options)
        elif self._browser_type == 'webkit':
            browser = await self._playwright.webkit.launch(**self._browser_options)
        else:
            raise ValueError(f'Invalid browser type: {self._browser_type}')

        return PlaywrightBrowserController(
            browser,
            max_open_pages_per_browser=self._max_open_pages_per_browser,
        )



from __future__ import annotations

import asyncio
import io
import os
from datetime import datetime, timezone
from logging import getLogger
from typing import TYPE_CHECKING, Any, AsyncContextManager

import aiofiles
import aioshutil
from aiofiles.os import makedirs
from typing_extensions import override

from crawlee._utils.crypto import crypto_random_object_id
from crawlee._utils.data_processing import maybe_parse_body, raise_on_duplicate_storage, raise_on_non_existing_storage
from crawlee._utils.file import determine_file_extension, force_remove, force_rename, is_file_or_bytes, json_dumps
from crawlee.base_storage_client import BaseKeyValueStoreClient
from crawlee.memory_storage_client._creation_management import (
    find_or_create_client_by_id_or_name_inner,
    persist_metadata_if_enabled,
)
from crawlee.models import (
    KeyValueStoreKeyInfo,
    KeyValueStoreListKeysPage,
    KeyValueStoreMetadata,
    KeyValueStoreRecord,
    KeyValueStoreRecordMetadata,
)
from crawlee.types import StorageTypes

if TYPE_CHECKING:
    from httpx import Response

    from crawlee.memory_storage_client import MemoryStorageClient

logger = getLogger(__name__)


class KeyValueStoreClient(BaseKeyValueStoreClient):
    """Subclient for manipulating a single key-value store."""

    def __init__(
        self,
        *,
        memory_storage_client: MemoryStorageClient,
        id: str | None = None,
        name: str | None = None,
        created_at: datetime | None = None,
        accessed_at: datetime | None = None,
        modified_at: datetime | None = None,
    ) -> None:
        self.id = id or crypto_random_object_id()
        self.name = name

        self._memory_storage_client = memory_storage_client
        self._created_at = created_at or datetime.now(timezone.utc)
        self._accessed_at = accessed_at or datetime.now(timezone.utc)
        self._modified_at = modified_at or datetime.now(timezone.utc)

        self.records: dict[str, KeyValueStoreRecord] = {}
        self.file_operation_lock = asyncio.Lock()

    @property
    def resource_info(self) -> KeyValueStoreMetadata:
        """Get the resource info for the key-value store client."""
        return KeyValueStoreMetadata(
            id=self.id,
            name=self.name,
            accessed_at=self._accessed_at,
            created_at=self._created_at,
            modified_at=self._modified_at,
            user_id='1',
        )

    @property
    def resource_directory(self) -> str:
        """Get the resource directory for the client."""
        return os.path.join(self._memory_storage_client.key_value_stores_directory, self.name or self.id)

    @override
    async def get(self) -> KeyValueStoreMetadata | None:
        found = find_or_create_client_by_id_or_name_inner(
            resource_client_class=KeyValueStoreClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if found:
            async with found.file_operation_lock:
                await found.update_timestamps(has_been_modified=False)
                return found.resource_info

        return None

    @override
    async def update(self, *, name: str | None = None) -> KeyValueStoreMetadata:
        # Check by id
        existing_store_by_id = find_or_create_client_by_id_or_name_inner(
            resource_client_class=KeyValueStoreClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if existing_store_by_id is None:
            raise_on_non_existing_storage(StorageTypes.KEY_VALUE_STORE, self.id)

        # Skip if no changes
        if name is None:
            return existing_store_by_id.resource_info

        async with existing_store_by_id.file_operation_lock:
            # Check that name is not in use already
            existing_store_by_name = next(
                (
                    store
                    for store in self._memory_storage_client.key_value_stores_handled
                    if store.name and store.name.lower() == name.lower()
                ),
                None,
            )

            if existing_store_by_name is not None:
                raise_on_duplicate_storage(StorageTypes.KEY_VALUE_STORE, 'name', name)

            previous_dir = existing_store_by_id.resource_directory
            existing_store_by_id.name = name

            await force_rename(previous_dir, existing_store_by_id.resource_directory)

            # Update timestamps
            await existing_store_by_id.update_timestamps(has_been_modified=True)

        return existing_store_by_id.resource_info

    @override
    async def delete(self) -> None:
        store = next(
            (store for store in self._memory_storage_client.key_value_stores_handled if store.id == self.id), None
        )

        if store is not None:
            async with store.file_operation_lock:
                self._memory_storage_client.key_value_stores_handled.remove(store)
                store.records.clear()

                if os.path.exists(store.resource_directory):
                    await aioshutil.rmtree(store.resource_directory)

    @override
    async def list_keys(
        self,
        *,
        limit: int = 1000,
        exclusive_start_key: str | None = None,
    ) -> KeyValueStoreListKeysPage:
        # Check by id
        existing_store_by_id = find_or_create_client_by_id_or_name_inner(
            resource_client_class=KeyValueStoreClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if existing_store_by_id is None:
            raise_on_non_existing_storage(StorageTypes.KEY_VALUE_STORE, self.id)

        items: list[KeyValueStoreKeyInfo] = []

        for record in existing_store_by_id.records.values():
            size = len(record.value)
            items.append(KeyValueStoreKeyInfo(key=record.key, size=size))

        if len(items) == 0:
            return KeyValueStoreListKeysPage(
                count=len(items),
                limit=limit,
                exclusive_start_key=exclusive_start_key,
                is_truncated=False,
                next_exclusive_start_key=None,
                items=items,
            )

        # Lexically sort to emulate the API
        items = sorted(items, key=lambda item: item.key)

        truncated_items = items
        if exclusive_start_key is not None:
            key_pos = next((idx for idx, item in enumerate(items) if item.key == exclusive_start_key), None)
            if key_pos is not None:
                truncated_items = items[(key_pos + 1) :]

        limited_items = truncated_items[:limit]

        last_item_in_store = items[-1]
        last_selected_item = limited_items[-1]
        is_last_selected_item_absolutely_last = last_item_in_store == last_selected_item
        next_exclusive_start_key = None if is_last_selected_item_absolutely_last else last_selected_item.key

        async with existing_store_by_id.file_operation_lock:
            await existing_store_by_id.update_timestamps(has_been_modified=False)

        return KeyValueStoreListKeysPage(
            count=len(items),
            limit=limit,
            exclusive_start_key=exclusive_start_key,
            is_truncated=not is_last_selected_item_absolutely_last,
            next_exclusive_start_key=next_exclusive_start_key,
            items=limited_items,
        )

    @override
    async def get_record(self, key: str) -> KeyValueStoreRecord | None:
        return await self._get_record_internal(key)

    @override
    async def get_record_as_bytes(self, key: str) -> KeyValueStoreRecord[bytes] | None:
        return await self._get_record_internal(key, as_bytes=True)

    @override
    async def stream_record(self, key: str) -> AsyncContextManager[KeyValueStoreRecord[Response] | None]:
        raise NotImplementedError('This method is not supported in memory storage.')

    @override
    async def set_record(self, key: str, value: Any, content_type: str | None = None) -> None:
        # Check by id
        existing_store_by_id = find_or_create_client_by_id_or_name_inner(
            resource_client_class=KeyValueStoreClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if existing_store_by_id is None:
            raise_on_non_existing_storage(StorageTypes.KEY_VALUE_STORE, self.id)

        if isinstance(value, io.IOBase):
            raise NotImplementedError('File-like values are not supported in local memory storage')

        if content_type is None:
            if is_file_or_bytes(value):
                content_type = 'application/octet-stream'
            elif isinstance(value, str):
                content_type = 'text/plain; charset=utf-8'
            else:
                content_type = 'application/json; charset=utf-8'

        if 'application/json' in content_type and not is_file_or_bytes(value) and not isinstance(value, str):
            s = await json_dumps(value)
            value = s.encode('utf-8')

        async with existing_store_by_id.file_operation_lock:
            await existing_store_by_id.update_timestamps(has_been_modified=True)
            record = KeyValueStoreRecord(key=key, value=value, content_type=content_type, filename=None)

            old_record = existing_store_by_id.records.get(key)
            existing_store_by_id.records[key] = record

            if self._memory_storage_client.persist_storage:
                record_filename = self._filename_from_record(record)
                record.filename = record_filename

                if old_record is not None and self._filename_from_record(old_record) != record_filename:
                    await existing_store_by_id.delete_persisted_record(old_record)

                await existing_store_by_id.persist_record(record)

    @override
    async def delete_record(self, key: str) -> None:
        # Check by id
        existing_store_by_id = find_or_create_client_by_id_or_name_inner(
            resource_client_class=KeyValueStoreClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if existing_store_by_id is None:
            raise_on_non_existing_storage(StorageTypes.KEY_VALUE_STORE, self.id)

        record = existing_store_by_id.records.get(key)

        if record is not None:
            async with existing_store_by_id.file_operation_lock:
                del existing_store_by_id.records[key]
                await existing_store_by_id.update_timestamps(has_been_modified=True)
                if self._memory_storage_client.persist_storage:
                    await existing_store_by_id.delete_persisted_record(record)

    async def persist_record(self, record: KeyValueStoreRecord) -> None:
        """Persist the specified record to the key-value store."""
        store_directory = self.resource_directory
        record_filename = self._filename_from_record(record)
        record.filename = record_filename
        record.content_type = record.content_type or 'application/octet-stream'

        # Ensure the directory for the entity exists
        await makedirs(store_directory, exist_ok=True)

        # Create files for the record
        record_path = os.path.join(store_directory, record_filename)
        record_metadata_path = os.path.join(store_directory, f'{record_filename}.__metadata__.json')

        # Convert to bytes if string
        if isinstance(record.value, str):
            record.value = record.value.encode('utf-8')

        async with aiofiles.open(record_path, mode='wb') as f:
            await f.write(record.value)

        if self._memory_storage_client.write_metadata:
            async with aiofiles.open(record_metadata_path, mode='wb') as f:
                record_metadata = KeyValueStoreRecordMetadata(key=record.key, content_type=record.content_type)
                await f.write(record_metadata.model_dump_json(indent=2).encode('utf-8'))

    async def delete_persisted_record(self, record: KeyValueStoreRecord) -> None:
        """Delete the specified record from the key-value store."""
        store_directory = self.resource_directory
        record_filename = self._filename_from_record(record)

        # Ensure the directory for the entity exists
        await makedirs(store_directory, exist_ok=True)

        # Create files for the record
        record_path = os.path.join(store_directory, record_filename)
        record_metadata_path = os.path.join(store_directory, record_filename + '.__metadata__.json')

        await force_remove(record_path)
        await force_remove(record_metadata_path)

    async def update_timestamps(self, *, has_been_modified: bool) -> None:
        """Update the timestamps of the key-value store."""
        self._accessed_at = datetime.now(timezone.utc)

        if has_been_modified:
            self._modified_at = datetime.now(timezone.utc)

        await persist_metadata_if_enabled(
            data=self.resource_info.model_dump(),
            entity_directory=self.resource_directory,
            write_metadata=self._memory_storage_client.write_metadata,
        )

    async def _get_record_internal(
        self,
        key: str,
        *,
        as_bytes: bool = False,
    ) -> KeyValueStoreRecord | None:
        # Check by id
        existing_store_by_id = find_or_create_client_by_id_or_name_inner(
            resource_client_class=KeyValueStoreClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if existing_store_by_id is None:
            raise_on_non_existing_storage(StorageTypes.KEY_VALUE_STORE, self.id)

        stored_record = existing_store_by_id.records.get(key)

        if stored_record is None:
            return None

        record = KeyValueStoreRecord(
            key=stored_record.key,
            value=stored_record.value,
            content_type=stored_record.content_type,
            filename=stored_record.filename,
        )

        if not as_bytes:
            try:
                record.value = maybe_parse_body(record.value, str(record.content_type))
            except ValueError:
                logger.exception('Error parsing key-value store record')

        async with existing_store_by_id.file_operation_lock:
            await existing_store_by_id.update_timestamps(has_been_modified=False)

        return record

    def _filename_from_record(self, record: KeyValueStoreRecord) -> str:
        if record.filename is not None:
            return record.filename

        if not record.content_type or record.content_type == 'application/octet-stream':
            return record.key

        extension = determine_file_extension(record.content_type)

        if record.key.endswith(f'.{extension}'):
            return record.key

        return f'{record.key}.{extension}'



from .base_dataset_client import BaseDatasetClient
from .base_dataset_collection_client import BaseDatasetCollectionClient
from .base_key_value_store_client import BaseKeyValueStoreClient
from .base_key_value_store_collection_client import BaseKeyValueStoreCollectionClient
from .base_request_queue_client import BaseRequestQueueClient
from .base_request_queue_collection_client import BaseRequestQueueCollectionClient
from .base_storage_client import BaseStorageClient

__all__ = [
    'BaseDatasetClient',
    'BaseDatasetCollectionClient',
    'BaseKeyValueStoreClient',
    'BaseKeyValueStoreCollectionClient',
    'BaseRequestQueueClient',
    'BaseRequestQueueCollectionClient',
    'BaseStorageClient',
]



from __future__ import annotations

import re
from base64 import b64encode
from hashlib import sha256
from logging import getLogger
from urllib.parse import parse_qsl, urlencode, urlparse

from crawlee._utils.crypto import compute_short_hash

logger = getLogger(__name__)


def unique_key_to_request_id(unique_key: str, *, request_id_length: int = 15) -> str:
    """Generate a deterministic request ID based on a unique key.

    Args:
        unique_key: The unique key to convert into a request ID.
        request_id_length: The length of the request ID.

    Returns:
        A URL-safe, truncated request ID based on the unique key.
    """
    # Encode the unique key and compute its SHA-256 hash
    hashed_key = sha256(unique_key.encode('utf-8')).digest()

    # Encode the hash in base64 and decode it to get a string
    base64_encoded = b64encode(hashed_key).decode('utf-8')

    # Remove characters that are not URL-safe ('+', '/', or '=')
    url_safe_key = re.sub(r'(\+|\/|=)', '', base64_encoded)

    # Truncate the key to the desired length
    return url_safe_key[:request_id_length]


def normalize_url(url: str, *, keep_url_fragment: bool = False) -> str:
    """Normalizes a URL.

    This function cleans and standardizes a URL by removing leading and trailing whitespaces,
    converting the scheme and netloc to lower case, stripping unwanted tracking parameters
    (specifically those beginning with 'utm_'), sorting the remaining query parameters alphabetically,
    and optionally retaining the URL fragment. The goal is to ensure that URLs that are functionally
    identical but differ in trivial ways (such as parameter order or casing) are treated as the same.

    Args:
        url: The URL to be normalized.
        keep_url_fragment: Flag to determine whether the fragment part of the URL should be retained.

    Returns:
        A string containing the normalized URL.
    """
    # Parse the URL
    parsed_url = urlparse(url.strip())
    search_params = dict(parse_qsl(parsed_url.query))  # Convert query to a dict

    # Remove any 'utm_' parameters
    search_params = {k: v for k, v in search_params.items() if not k.startswith('utm_')}

    # Construct the new query string
    sorted_keys = sorted(search_params.keys())
    sorted_query = urlencode([(k, search_params[k]) for k in sorted_keys])

    # Construct the final URL
    new_url = (
        parsed_url._replace(
            query=sorted_query,
            scheme=parsed_url.scheme,
            netloc=parsed_url.netloc,
            path=parsed_url.path.rstrip('/'),
        )
        .geturl()
        .lower()
    )

    # Retain the URL fragment if required
    if not keep_url_fragment:
        new_url = new_url.split('#')[0]

    return new_url


def compute_unique_key(
    url: str,
    method: str = 'GET',
    payload: bytes | None = None,
    *,
    keep_url_fragment: bool = False,
    use_extended_unique_key: bool = False,
) -> str:
    """Computes a unique key for caching & deduplication of requests.

    This function computes a unique key by normalizing the provided URL and method.
    If 'use_extended_unique_key' is True and a payload is provided, the payload is hashed and
    included in the key. Otherwise, the unique key is just the normalized URL.

    Args:
        url: The request URL.
        method: The HTTP method, defaults to 'GET'.
        payload: The request payload, defaults to None.
        keep_url_fragment: A flag indicating whether to keep the URL fragment, defaults to False.
        use_extended_unique_key: A flag indicating whether to include a hashed payload in the key, defaults to False.

    Returns:
        A string representing the unique key for the request.
    """
    # Normalize the URL and method.
    try:
        normalized_url = normalize_url(url, keep_url_fragment=keep_url_fragment)
    except Exception as exc:
        logger.warning(f'Failed to normalize URL: {exc}')
        normalized_url = url

    normalized_method = method.upper()

    # Compute and return the extended unique key if required.
    if use_extended_unique_key:
        payload_hash = compute_short_hash(payload) if payload else ''
        return f'{normalized_method}({payload_hash}):{normalized_url}'

    # Log information if there is a non-GET request with a payload.
    if normalized_method != 'GET' and payload:
        logger.info(
            f'We have encountered a {normalized_method} Request with a payload. This is fine. Just letting you know '
            'that if your requests point to the same URL and differ only in method and payload, you should consider '
            'using the "use_extended_unique_key" option.'
        )

    # Return the normalized URL as the unique key.
    return normalized_url



try:
    from .beautifulsoup_crawler import BeautifulSoupCrawler
    from .types import BeautifulSoupCrawlingContext
except ImportError as exc:
    raise ImportError(
        'To import anything from this subpacakge, you need to install the "beautifulsoup" extra. '
        'For example, if you use pip, run "pip install crawlee[beautifulsoup]".',
    ) from exc

__all__ = ['BeautifulSoupCrawler', 'BeautifulSoupCrawlingContext']



from __future__ import annotations

import asyncio
import json
import os
from datetime import datetime, timezone
from decimal import Decimal
from logging import getLogger
from typing import TYPE_CHECKING

import aiofiles
import aioshutil
from aiofiles.os import makedirs
from sortedcollections import ValueSortedDict  # type: ignore
from typing_extensions import override

from crawlee._utils.crypto import crypto_random_object_id
from crawlee._utils.data_processing import (
    filter_out_none_values_recursively,
    raise_on_duplicate_storage,
    raise_on_non_existing_storage,
)
from crawlee._utils.file import force_remove, force_rename, json_dumps
from crawlee._utils.requests import unique_key_to_request_id
from crawlee.base_storage_client import BaseRequestQueueClient
from crawlee.memory_storage_client._creation_management import (
    find_or_create_client_by_id_or_name_inner,
    persist_metadata_if_enabled,
)
from crawlee.models import (
    BatchRequestsOperationResponse,
    ProcessedRequest,
    ProlongRequestLockResponse,
    Request,
    RequestListResponse,
    RequestQueueHead,
    RequestQueueHeadWithLocks,
    RequestQueueMetadata,
    UnprocessedRequest,
)
from crawlee.types import StorageTypes

if TYPE_CHECKING:
    from collections.abc import Sequence

    from crawlee.memory_storage_client import MemoryStorageClient

logger = getLogger(__name__)


class RequestQueueClient(BaseRequestQueueClient):
    """Subclient for manipulating a single request queue."""

    def __init__(
        self,
        *,
        memory_storage_client: MemoryStorageClient,
        id: str | None = None,
        name: str | None = None,
        created_at: datetime | None = None,
        accessed_at: datetime | None = None,
        modified_at: datetime | None = None,
        handled_request_count: int = 0,
        pending_request_count: int = 0,
    ) -> None:
        self._memory_storage_client = memory_storage_client
        self.id = id or crypto_random_object_id()
        self.name = name
        self._created_at = created_at or datetime.now(timezone.utc)
        self._accessed_at = accessed_at or datetime.now(timezone.utc)
        self._modified_at = modified_at or datetime.now(timezone.utc)
        self.handled_request_count = handled_request_count
        self.pending_request_count = pending_request_count

        self.requests = ValueSortedDict(lambda request: request.order_no or -float('inf'))
        self.file_operation_lock = asyncio.Lock()
        self._last_used_timestamp = Decimal(0.0)

    @property
    def resource_info(self) -> RequestQueueMetadata:
        """Get the resource info for the request queue client."""
        return RequestQueueMetadata(
            id=self.id,
            name=self.name,
            accessed_at=self._accessed_at,
            created_at=self._created_at,
            modified_at=self._modified_at,
            had_multiple_clients=False,
            handled_request_count=self.handled_request_count,
            pending_request_count=self.pending_request_count,
            stats={},
            total_request_count=len(self.requests),
            user_id='1',
            resource_directory=self.resource_directory,
        )

    @property
    def resource_directory(self) -> str:
        """Get the resource directory for the client."""
        return os.path.join(self._memory_storage_client.request_queues_directory, self.name or self.id)

    @override
    async def get(self) -> RequestQueueMetadata | None:
        found = find_or_create_client_by_id_or_name_inner(
            resource_client_class=RequestQueueClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if found:
            async with found.file_operation_lock:
                await found.update_timestamps(has_been_modified=False)
                return found.resource_info

        return None

    @override
    async def update(self, *, name: str | None = None) -> RequestQueueMetadata:
        # Check by id
        existing_queue_by_id = find_or_create_client_by_id_or_name_inner(
            resource_client_class=RequestQueueClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if existing_queue_by_id is None:
            raise_on_non_existing_storage(StorageTypes.REQUEST_QUEUE, self.id)

        # Skip if no changes
        if name is None:
            return existing_queue_by_id.resource_info

        async with existing_queue_by_id.file_operation_lock:
            # Check that name is not in use already
            existing_queue_by_name = next(
                (
                    queue
                    for queue in self._memory_storage_client.request_queues_handled
                    if queue.name and queue.name.lower() == name.lower()
                ),
                None,
            )

            if existing_queue_by_name is not None:
                raise_on_duplicate_storage(StorageTypes.REQUEST_QUEUE, 'name', name)

            previous_dir = existing_queue_by_id.resource_directory
            existing_queue_by_id.name = name

            await force_rename(previous_dir, existing_queue_by_id.resource_directory)

            # Update timestamps
            await existing_queue_by_id.update_timestamps(has_been_modified=True)

            return existing_queue_by_id.resource_info

    @override
    async def delete(self) -> None:
        queue = next(
            (queue for queue in self._memory_storage_client.request_queues_handled if queue.id == self.id),
            None,
        )

        if queue is not None:
            async with queue.file_operation_lock:
                self._memory_storage_client.request_queues_handled.remove(queue)
                queue.pending_request_count = 0
                queue.handled_request_count = 0
                queue.requests.clear()

                if os.path.exists(queue.resource_directory):
                    await aioshutil.rmtree(queue.resource_directory)

    @override
    async def list_head(self, *, limit: int | None = None) -> RequestQueueHead:
        existing_queue_by_id = find_or_create_client_by_id_or_name_inner(
            resource_client_class=RequestQueueClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if existing_queue_by_id is None:
            raise_on_non_existing_storage(StorageTypes.REQUEST_QUEUE, self.id)

        async with existing_queue_by_id.file_operation_lock:
            await existing_queue_by_id.update_timestamps(has_been_modified=False)

            requests: list[Request] = []

            # Iterate all requests in the queue which have sorted key larger than infinity, which means
            # `order_no` is not `None`. This will iterate them in order of `order_no`.
            for request_key in existing_queue_by_id.requests.irange_key(
                min_key=-float('inf'),
                inclusive=(False, True),
            ):
                if len(requests) == limit:
                    break

                request = existing_queue_by_id.requests.get(request_key)

                # Check that the request still exists and was not handled,
                # in case something deleted it or marked it as handled concurrenctly
                if request and request.order_no:
                    requests.append(request)

            items = [request for item in requests if (request := self._json_to_request(item.json_))]

            return RequestQueueHead(
                limit=limit,
                had_multiple_clients=False,
                queue_modified_at=existing_queue_by_id._modified_at,  # noqa: SLF001
                items=items,
            )

    @override
    async def list_and_lock_head(self, *, lock_secs: int, limit: int | None = None) -> RequestQueueHeadWithLocks:
        raise NotImplementedError('This method is not supported in memory storage.')

    @override
    async def add_request(
        self,
        request: Request,
        *,
        forefront: bool = False,
    ) -> ProcessedRequest:
        existing_queue_by_id = find_or_create_client_by_id_or_name_inner(
            resource_client_class=RequestQueueClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if existing_queue_by_id is None:
            raise_on_non_existing_storage(StorageTypes.REQUEST_QUEUE, self.id)

        request_model = await self._create_internal_request(request, forefront)

        async with existing_queue_by_id.file_operation_lock:
            existing_request_with_id = existing_queue_by_id.requests.get(request_model.id)

            # We already have the request present, so we return information about it
            if existing_request_with_id is not None:
                await existing_queue_by_id.update_timestamps(has_been_modified=False)

                return ProcessedRequest(
                    id=request_model.id,
                    unique_key=request_model.unique_key,
                    was_already_present=True,
                    was_already_handled=existing_request_with_id.order_no is None,
                )

            existing_queue_by_id.requests[request_model.id] = request_model
            if request_model.order_no is None:
                existing_queue_by_id.handled_request_count += 1
            else:
                existing_queue_by_id.pending_request_count += 1
            await existing_queue_by_id.update_timestamps(has_been_modified=True)
            await self._persist_single_request_to_storage(
                request=request_model,
                entity_directory=existing_queue_by_id.resource_directory,
                persist_storage=self._memory_storage_client.persist_storage,
            )

            # We return wasAlreadyHandled is false even though the request may have been added as handled,
            # because that's how API behaves.
            return ProcessedRequest(
                id=request_model.id,
                unique_key=request_model.unique_key,
                was_already_present=False,
                was_already_handled=False,
            )

    @override
    async def get_request(self, request_id: str) -> Request | None:
        existing_queue_by_id = find_or_create_client_by_id_or_name_inner(
            resource_client_class=RequestQueueClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if existing_queue_by_id is None:
            raise_on_non_existing_storage(StorageTypes.REQUEST_QUEUE, self.id)

        async with existing_queue_by_id.file_operation_lock:
            await existing_queue_by_id.update_timestamps(has_been_modified=False)

            request: Request = existing_queue_by_id.requests.get(request_id)
            return self._json_to_request(request.json_ if request is not None else None)

    @override
    async def update_request(
        self,
        request: Request,
        *,
        forefront: bool = False,
    ) -> ProcessedRequest:
        existing_queue_by_id = find_or_create_client_by_id_or_name_inner(
            resource_client_class=RequestQueueClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if existing_queue_by_id is None:
            raise_on_non_existing_storage(StorageTypes.REQUEST_QUEUE, self.id)

        request_model = await self._create_internal_request(request, forefront)

        # First we need to check the existing request to be able to return information about its handled state.
        existing_request = existing_queue_by_id.requests.get(request_model.id)

        # Undefined means that the request is not present in the queue.
        # We need to insert it, to behave the same as API.
        if existing_request is None:
            return await self.add_request(request, forefront=forefront)

        async with existing_queue_by_id.file_operation_lock:
            # When updating the request, we need to make sure that
            # the handled counts are updated correctly in all cases.
            existing_queue_by_id.requests[request_model.id] = request_model

            pending_count_adjustment = 0
            is_request_handled_state_changing = not isinstance(existing_request.order_no, type(request_model.order_no))
            request_was_handled_before_update = existing_request.order_no is None

            # We add 1 pending request if previous state was handled
            if is_request_handled_state_changing:
                pending_count_adjustment = 1 if request_was_handled_before_update else -1

            existing_queue_by_id.pending_request_count += pending_count_adjustment
            existing_queue_by_id.handled_request_count -= pending_count_adjustment
            await existing_queue_by_id.update_timestamps(has_been_modified=True)
            await self._persist_single_request_to_storage(
                request=request_model,
                entity_directory=existing_queue_by_id.resource_directory,
                persist_storage=self._memory_storage_client.persist_storage,
            )

            return ProcessedRequest(
                id=request_model.id,
                unique_key=request_model.unique_key,
                was_already_present=True,
                was_already_handled=request_was_handled_before_update,
            )

    @override
    async def delete_request(self, request_id: str) -> None:
        existing_queue_by_id = find_or_create_client_by_id_or_name_inner(
            resource_client_class=RequestQueueClient,
            memory_storage_client=self._memory_storage_client,
            id=self.id,
            name=self.name,
        )

        if existing_queue_by_id is None:
            raise_on_non_existing_storage(StorageTypes.REQUEST_QUEUE, self.id)

        async with existing_queue_by_id.file_operation_lock:
            request = existing_queue_by_id.requests.get(request_id)

            if request:
                del existing_queue_by_id.requests[request_id]
                if request.order_no is None:
                    existing_queue_by_id.handled_request_count -= 1
                else:
                    existing_queue_by_id.pending_request_count -= 1
                await existing_queue_by_id.update_timestamps(has_been_modified=True)
                await self._delete_request_file_from_storage(
                    entity_directory=existing_queue_by_id.resource_directory,
                    request_id=request_id,
                )

    @override
    async def prolong_request_lock(
        self,
        request_id: str,
        *,
        forefront: bool = False,
        lock_secs: int,
    ) -> ProlongRequestLockResponse:
        raise NotImplementedError('This method is not supported in memory storage.')

    @override
    async def delete_request_lock(
        self,
        request_id: str,
        *,
        forefront: bool = False,
    ) -> None:
        raise NotImplementedError('This method is not supported in memory storage.')

    @override
    async def batch_add_requests(
        self,
        requests: Sequence[Request],
        *,
        forefront: bool = False,
    ) -> BatchRequestsOperationResponse:
        processed_requests = list[ProcessedRequest]()
        unprocessed_requests = list[UnprocessedRequest]()

        for request in requests:
            try:
                processed_request = await self.add_request(request, forefront=forefront)
                processed_requests.append(
                    ProcessedRequest(
                        id=processed_request.id,
                        unique_key=processed_request.unique_key,
                        was_already_present=processed_request.was_already_present,
                        was_already_handled=processed_request.was_already_handled,
                    )
                )
            except Exception as exc:  # noqa: PERF203
                logger.warning(f'Error adding request to the queue: {exc}')
                unprocessed_requests.append(
                    UnprocessedRequest(
                        unique_key=request.unique_key,
                        url=request.url,
                        method=request.method,
                    )
                )

        return BatchRequestsOperationResponse(
            processed_requests=processed_requests,
            unprocessed_requests=unprocessed_requests,
        )

    @override
    async def batch_delete_requests(self, requests: list[Request]) -> BatchRequestsOperationResponse:
        raise NotImplementedError('This method is not supported in memory storage.')

    @override
    async def list_requests(
        self,
        *,
        limit: int | None = None,
        exclusive_start_id: str | None = None,
    ) -> RequestListResponse:
        raise NotImplementedError('This method is not supported in memory storage.')

    async def update_timestamps(self, *, has_been_modified: bool) -> None:
        """Update the timestamps of the request queue."""
        self._accessed_at = datetime.now(timezone.utc)

        if has_been_modified:
            self._modified_at = datetime.now(timezone.utc)

        await persist_metadata_if_enabled(
            data=self.resource_info.model_dump(),
            entity_directory=self.resource_directory,
            write_metadata=self._memory_storage_client.write_metadata,
        )

    async def _persist_single_request_to_storage(
        self,
        *,
        request: Request,
        entity_directory: str,
        persist_storage: bool,
    ) -> None:
        """Updates or writes a single request item to the disk.

        This function writes a given request dictionary to a JSON file, named after the request's ID,
        within a specified directory. The writing process is skipped if `persist_storage` is False.
        Before writing, it ensures that the target directory exists, creating it if necessary.

        Args:
            request: The dictionary containing the request data.
            entity_directory: The directory path where the request file should be stored.
            persist_storage: A boolean flag indicating whether the request should be persisted to the disk.
        """
        # Skip writing files to the disk if the client has the option set to false
        if not persist_storage:
            return

        # Ensure the directory for the entity exists
        await makedirs(entity_directory, exist_ok=True)

        # Write the request to the file
        file_path = os.path.join(entity_directory, f'{request.id}.json')
        async with aiofiles.open(file_path, mode='wb') as f:
            s = await json_dumps(request.model_dump())
            await f.write(s.encode('utf-8'))

    async def _delete_request_file_from_storage(self, *, request_id: str, entity_directory: str) -> None:
        """Deletes a specific request item from the disk.

        This function removes a file representing a request, identified by the request's ID, from a
        specified directory. Before attempting to remove the file, it ensures that the target directory
        exists, creating it if necessary.

        Args:
            request_id: The identifier of the request to be deleted.
            entity_directory: The directory path where the request file is stored.
        """
        # Ensure the directory for the entity exists
        await makedirs(entity_directory, exist_ok=True)

        file_path = os.path.join(entity_directory, f'{request_id}.json')
        await force_remove(file_path)

    def _json_to_request(self, request_json: str | None) -> Request | None:
        if request_json is None:
            return None
        request_dict = filter_out_none_values_recursively(json.loads(request_json))
        if request_dict is None:
            return None
        return Request.model_validate(request_dict)

    async def _create_internal_request(self, request: Request, forefront: bool | None) -> Request:
        order_no = self._calculate_order_no(request, forefront)
        id = unique_key_to_request_id(request.unique_key)

        if request.id is not None and request.id != id:
            raise ValueError('Request ID does not match its unique_key.')

        json_request = await json_dumps({**(request.model_dump()), 'id': id})
        return Request(
            url=request.url,
            unique_key=request.unique_key,
            id=id,
            method=request.method,
            retry_count=request.retry_count,
            order_no=order_no,
            json_=json_request,
            user_data={},
        )

    def _calculate_order_no(self, request: Request, forefront: bool | None) -> Decimal | None:
        if request.handled_at is not None:
            return None

        # Get the current timestamp in milliseconds
        timestamp = Decimal(datetime.now(timezone.utc).timestamp()) * 1000
        timestamp = round(timestamp, 6)

        # Make sure that this timestamp was not used yet, so that we have unique orderNos
        if timestamp <= self._last_used_timestamp:
            timestamp = self._last_used_timestamp + Decimal(0.000001)

        self._last_used_timestamp = timestamp

        return -timestamp if forefront else timestamp



from __future__ import annotations

import time
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Iterator


@dataclass
class TimerResult:
    wall: float | None = None
    cpu: float | None = None


@contextmanager
def measure_time() -> Iterator[TimerResult]:
    """Measure the execution time (wall-clock and CPU) between the start and end of the with-block."""
    result = TimerResult()
    before_wall = time.monotonic()
    before_cpu = time.thread_time()

    try:
        yield result
    finally:
        after_wall = time.monotonic()
        after_cpu = time.thread_time()
        result.wall = after_wall - before_wall
        result.cpu = after_cpu - before_cpu



from __future__ import annotations

from typing import TYPE_CHECKING, Any, AsyncIterator, TypeVar, overload

from typing_extensions import override

from crawlee.models import KeyValueStoreKeyInfo, KeyValueStoreMetadata
from crawlee.storages.base_storage import BaseStorage

if TYPE_CHECKING:
    from crawlee.base_storage_client import BaseStorageClient
    from crawlee.configuration import Configuration

T = TypeVar('T')


class KeyValueStore(BaseStorage):
    """Represents a key-value based storage for reading data records or files.

    Each record is identified by a unique key and associated with a MIME content type. This class is used within
    crawler runs to store inputs and outputs, typically in JSON format, but supports other types as well.

    The data can be stored on a local filesystem or in the cloud, determined by the `CRAWLEE_STORAGE_DIR`
    environment variable.

    By default, data is stored in `{CRAWLEE_STORAGE_DIR}/key_value_stores/{STORE_ID}/{INDEX}.{EXT}`, where
    `{STORE_ID}` is either "default" or specified by `CRAWLEE_DEFAULT_KEY_VALUE_STORE_ID`, `{KEY}` is the record key,
    and `{EXT}` is the MIME type.

    To open a key-value store, use the class method `open`, providing either an `id` or `name` along with optional
    `config`. If neither is provided, the default store for the crawler run is used. Opening a non-existent store by
    `id` raises an error, while a non-existent store by `name` is created.

    Usage:
        kvs = await KeyValueStore.open(id='my_kvs_id')
    """

    def __init__(
        self,
        id: str,
        name: str | None,
        configuration: Configuration,
        client: BaseStorageClient,
    ) -> None:
        self._id = id
        self._name = name
        self._configuration = configuration

        # Get resource clients from storage client
        self._resource_client = client.key_value_store(self._id)

    @override
    @property
    def id(self) -> str:
        return self._id

    @override
    @property
    def name(self) -> str | None:
        return self._name

    async def get_info(self) -> KeyValueStoreMetadata | None:
        """Get an object containing general information about the key value store."""
        return await self._resource_client.get()

    @override
    @classmethod
    async def open(
        cls,
        *,
        id: str | None = None,
        name: str | None = None,
        configuration: Configuration | None = None,
    ) -> KeyValueStore:
        from crawlee.storages._creation_management import open_storage

        return await open_storage(
            storage_class=cls,
            id=id,
            name=name,
            configuration=configuration,
        )

    @override
    async def drop(self) -> None:
        from crawlee.storages._creation_management import remove_storage_from_cache

        await self._resource_client.delete()
        remove_storage_from_cache(storage_class=self.__class__, id=self._id, name=self._name)

    @overload
    async def get_value(self, key: str) -> Any: ...

    @overload
    async def get_value(self, key: str, default_value: T) -> T: ...

    @overload
    async def get_value(self, key: str, default_value: T | None = None) -> T | None: ...

    async def get_value(self, key: str, default_value: T | None = None) -> T | None:
        """Get a value from the KVS.

        Args:
            key: Key of the record to retrieve.
            default_value: Default value returned in case the record does not exist.

        Returns:
            The value associated with the given key. `default_value` is used in case the record does not exist.
        """
        record = await self._resource_client.get_record(key)
        return record.value if record else default_value

    async def iterate_keys(self, exclusive_start_key: str | None = None) -> AsyncIterator[KeyValueStoreKeyInfo]:
        """Iterate over the existing keys in the KVS.

        Args:
            exclusive_start_key: Key to start the iteration from.

        Yields:
            Information about the key.
        """
        while True:
            list_keys = await self._resource_client.list_keys(exclusive_start_key=exclusive_start_key)
            for item in list_keys.items:
                yield KeyValueStoreKeyInfo(key=item.key, size=item.size)

            if not list_keys.is_truncated:
                break
            exclusive_start_key = list_keys.next_exclusive_start_key

    async def set_value(
        self,
        key: str,
        value: Any,
        content_type: str | None = None,
    ) -> None:
        """Set a value in the KVS.

        Args:
            key: Key of the record to set.
            value: Value to set. If `None`, the record is deleted.
            content_type: Content type of the record.
        """
        if value is None:
            return await self._resource_client.delete_record(key)

        return await self._resource_client.set_record(key, value, content_type)



# Inspiration: https://github.com/apify/crawlee/blob/v3.7.3/packages/basic-crawler/src/internals/basic-crawler.ts
from __future__ import annotations

import asyncio
import logging
import signal
import sys
import tempfile
from asyncio import CancelledError
from collections.abc import AsyncGenerator, Awaitable, Sequence
from contextlib import AsyncExitStack, suppress
from datetime import timedelta
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, AsyncContextManager, Callable, Generic, Literal, Union, cast

import httpx
from tldextract import TLDExtract
from typing_extensions import NotRequired, TypedDict, TypeVar, Unpack, assert_never

from crawlee import Glob
from crawlee._utils.wait import wait_for
from crawlee.autoscaling import AutoscaledPool, ConcurrencySettings
from crawlee.autoscaling.snapshotter import Snapshotter
from crawlee.autoscaling.system_status import SystemStatus
from crawlee.basic_crawler.context_pipeline import ContextPipeline
from crawlee.basic_crawler.errors import (
    ContextPipelineInitializationError,
    ContextPipelineInterruptedError,
    RequestHandlerError,
    SessionError,
    UserDefinedErrorHandlerError,
)
from crawlee.basic_crawler.router import Router
from crawlee.basic_crawler.types import BasicCrawlingContext, RequestHandlerRunResult, SendRequestFunction
from crawlee.configuration import Configuration
from crawlee.enqueue_strategy import EnqueueStrategy
from crawlee.events import LocalEventManager
from crawlee.http_clients import HttpxClient
from crawlee.log_config import CrawleeLogFormatter
from crawlee.models import BaseRequestData, DatasetItemsListPage, Request, RequestState
from crawlee.sessions import SessionPool
from crawlee.statistics import Statistics
from crawlee.storages import Dataset, KeyValueStore, RequestQueue

if TYPE_CHECKING:
    import re

    from crawlee.http_clients import BaseHttpClient, HttpResponse
    from crawlee.proxy_configuration import ProxyConfiguration, ProxyInfo
    from crawlee.sessions import Session
    from crawlee.statistics import FinalStatistics, StatisticsState
    from crawlee.storages.dataset import GetDataKwargs, PushDataKwargs
    from crawlee.storages.request_provider import RequestProvider
    from crawlee.types import JSONSerializable

TCrawlingContext = TypeVar('TCrawlingContext', bound=BasicCrawlingContext, default=BasicCrawlingContext)
ErrorHandler = Callable[[TCrawlingContext, Exception], Awaitable[Union[Request, None]]]
FailedRequestHandler = Callable[[TCrawlingContext, Exception], Awaitable[None]]


class BasicCrawlerOptions(TypedDict, Generic[TCrawlingContext]):
    """Copy of the parameter types of `BasicCrawler.__init__` meant for typing forwarded __init__ args in subclasses."""

    request_provider: NotRequired[RequestProvider]
    request_handler: NotRequired[Callable[[TCrawlingContext], Awaitable[None]]]
    http_client: NotRequired[BaseHttpClient]
    concurrency_settings: NotRequired[ConcurrencySettings]
    max_request_retries: NotRequired[int]
    max_requests_per_crawl: NotRequired[int | None]
    max_session_rotations: NotRequired[int]
    configuration: NotRequired[Configuration]
    request_handler_timeout: NotRequired[timedelta]
    session_pool: NotRequired[SessionPool]
    use_session_pool: NotRequired[bool]
    retry_on_blocked: NotRequired[bool]
    proxy_configuration: NotRequired[ProxyConfiguration]
    statistics: NotRequired[Statistics[StatisticsState]]
    configure_logging: NotRequired[bool]
    _context_pipeline: NotRequired[ContextPipeline[TCrawlingContext]]
    _additional_context_managers: NotRequired[Sequence[AsyncContextManager]]
    _logger: NotRequired[logging.Logger]


class BasicCrawler(Generic[TCrawlingContext]):
    """Provides a simple framework for parallel crawling of web pages.

    The URLs to crawl are fed either from a static list of URLs or from a dynamic queue of URLs enabling recursive
    crawling of websites.

    `BasicCrawler` is a low-level tool that requires the user to implement the page download and data extraction
    functionality themselves. If we want a crawler that already facilitates this functionality, we should consider using
    one of its subclasses.
    """

    def __init__(
        self,
        *,
        request_provider: RequestProvider | None = None,
        request_handler: Callable[[TCrawlingContext], Awaitable[None]] | None = None,
        http_client: BaseHttpClient | None = None,
        concurrency_settings: ConcurrencySettings | None = None,
        max_request_retries: int = 3,
        max_requests_per_crawl: int | None = None,
        max_session_rotations: int = 10,
        configuration: Configuration | None = None,
        request_handler_timeout: timedelta = timedelta(minutes=1),
        session_pool: SessionPool | None = None,
        use_session_pool: bool = True,
        retry_on_blocked: bool = True,
        proxy_configuration: ProxyConfiguration | None = None,
        statistics: Statistics | None = None,
        configure_logging: bool = True,
        _context_pipeline: ContextPipeline[TCrawlingContext] | None = None,
        _additional_context_managers: Sequence[AsyncContextManager] | None = None,
        _logger: logging.Logger | None = None,
    ) -> None:
        """Initialize the BasicCrawler.

        Args:
            request_provider: Provides requests to be processed
            request_handler: A callable to which request handling is delegated
            http_client: HTTP client to be used for `BasicCrawlingContext.send_request` and HTTP-only crawling.
            concurrency_settings: Allows fine-tuning concurrency levels
            max_request_retries: Maximum amount of attempts at processing a request
            max_requests_per_crawl: Maximum number of pages that the crawler will open. The crawl will stop when
                the limit is reached. It is recommended to set this value in order to prevent infinite loops in
                misconfigured crawlers. None means no limit. Due to concurrency_settings, the actual number of pages
                visited may slightly exceed this value.
            max_session_rotations: Maximum number of session rotations per request.
                The crawler will automatically rotate the session in case of a proxy error or if it gets blocked by
                the website.
            configuration: Crawler configuration
            request_handler_timeout: How long is a single request handler allowed to run
            use_session_pool: Enables using the session pool for crawling
            session_pool: A preconfigured `SessionPool` instance if you wish to use non-default configuration
            retry_on_blocked: If set to True, the crawler will try to automatically bypass any detected bot protection
            proxy_configuration: A HTTP proxy configuration to be used for making requests
            statistics: A preconfigured `Statistics` instance if you wish to use non-default configuration
            configure_logging: If set to True, the crawler will configure the logging infrastructure
            _context_pipeline: Allows extending the request lifecycle and modifying the crawling context.
                This parameter is meant to be used by child classes, not when BasicCrawler is instantiated directly.
            _additional_context_managers: Additional context managers to be used in the crawler lifecycle.
            _logger: A logger instance passed from a child class to ensure consistent labels
        """
        self._router: Router[TCrawlingContext] | None = None

        if isinstance(cast(Router, request_handler), Router):
            self._router = cast(Router[TCrawlingContext], request_handler)
        elif request_handler is not None:
            self._router = None
            self.router.default_handler(request_handler)

        self._http_client = http_client or HttpxClient()

        self._context_pipeline = (_context_pipeline or ContextPipeline()).compose(self._check_url_after_redirects)

        self._error_handler: ErrorHandler[TCrawlingContext] | None = None
        self._failed_request_handler: FailedRequestHandler[TCrawlingContext] | None = None

        self._max_request_retries = max_request_retries
        self._max_requests_per_crawl = max_requests_per_crawl
        self._max_session_rotations = max_session_rotations

        self._request_provider = request_provider
        self._configuration = configuration or Configuration.get_global_configuration()

        self._request_handler_timeout = request_handler_timeout
        self._internal_timeout = (
            self._configuration.internal_timeout
            if self._configuration.internal_timeout is not None
            else max(2 * request_handler_timeout, timedelta(minutes=5))
        )

        self._tld_extractor = TLDExtract(cache_dir=tempfile.TemporaryDirectory().name)

        self._event_manager = LocalEventManager()  # TODO: switch based on configuration
        # https://github.com/apify/crawlee-py/issues/83
        self._snapshotter = Snapshotter(self._event_manager)
        self._pool = AutoscaledPool(
            system_status=SystemStatus(self._snapshotter),
            is_finished_function=self.__is_finished_function,
            is_task_ready_function=self.__is_task_ready_function,
            run_task_function=self.__run_task_function,
            concurrency_settings=concurrency_settings,
        )

        self._use_session_pool = use_session_pool
        self._session_pool = session_pool or SessionPool()

        self._retry_on_blocked = retry_on_blocked

        if configure_logging:
            handler = logging.StreamHandler()
            handler.setFormatter(CrawleeLogFormatter())

            root_logger = logging.getLogger()

            for old_handler in root_logger.handlers[:]:
                root_logger.removeHandler(old_handler)

            root_logger.addHandler(handler)
            root_logger.setLevel(logging.INFO if not sys.flags.dev_mode else logging.DEBUG)

            # Silence HTTPX logger
            httpx_logger = logging.getLogger('httpx')
            httpx_logger.setLevel(logging.WARNING if not sys.flags.dev_mode else logging.INFO)

        if not _logger:
            _logger = logging.getLogger(__name__)

        self._logger = _logger

        self._proxy_configuration = proxy_configuration
        self._statistics = statistics or Statistics(
            event_manager=self._event_manager,
            log_message=f'{self._logger.name} request statistics',
        )
        self._additional_context_managers = _additional_context_managers or []

        self._running = False
        self._has_finished_before = False

    @property
    def log(self) -> logging.Logger:
        """The logger used by the crawler."""
        return self._logger

    @property
    def router(self) -> Router[TCrawlingContext]:
        """The router used to handle each individual crawling request."""
        if self._router is None:
            self._router = Router[TCrawlingContext]()

        return self._router

    @router.setter
    def router(self, router: Router[TCrawlingContext]) -> None:
        if self._router is not None:
            raise RuntimeError('A router is already set')

        self._router = router

    @property
    def statistics(self) -> Statistics[StatisticsState]:
        """Statistics about the current (or last) crawler run."""
        return self._statistics

    @property
    def _max_requests_count_exceeded(self) -> bool:
        """Whether the maximum number of requests to crawl has been reached."""
        if self._max_requests_per_crawl is None:
            return False

        return self._statistics.state.requests_finished >= self._max_requests_per_crawl

    async def _get_session(self) -> Session | None:
        """If session pool is being used, try to take a session from it."""
        if not self._use_session_pool:
            return None

        return await wait_for(
            self._session_pool.get_session,
            timeout=self._internal_timeout,
            timeout_message='Fetching a session from the pool timed out after '
            f'{self._internal_timeout.total_seconds()} seconds',
            max_retries=3,
            logger=self._logger,
        )

    async def _get_proxy_info(self, request: Request, session: Session | None) -> ProxyInfo | None:
        """Retrieve a new ProxyInfo object based on crawler configuration and the current request and session."""
        if not self._proxy_configuration:
            return None

        return await self._proxy_configuration.new_proxy_info(
            session_id=session.id if session else None,
            request=request,
            proxy_tier=None,
        )

    async def get_request_provider(
        self,
        *,
        id: str | None = None,
        name: str | None = None,
    ) -> RequestProvider:
        """Return the configured request provider. If none is configured, open and return the default request queue."""
        if not self._request_provider:
            self._request_provider = await RequestQueue.open(id=id, name=name)

        return self._request_provider

    async def get_dataset(
        self,
        *,
        id: str | None = None,
        name: str | None = None,
    ) -> Dataset:
        """Return the dataset with the given ID or name. If none is provided, return the default dataset."""
        return await Dataset.open(id=id, name=name)

    async def get_key_value_store(
        self,
        *,
        id: str | None = None,
        name: str | None = None,
    ) -> KeyValueStore:
        """Return the key-value store with the given ID or name. If none is provided, return the default KVS."""
        return await KeyValueStore.open(id=id, name=name)

    def error_handler(self, handler: ErrorHandler[TCrawlingContext]) -> ErrorHandler[TCrawlingContext]:
        """Decorator for configuring an error handler (called after a request handler error and before retrying)."""
        self._error_handler = handler
        return handler

    def failed_request_handler(
        self, handler: FailedRequestHandler[TCrawlingContext]
    ) -> FailedRequestHandler[TCrawlingContext]:
        """Decorator for configuring a failed request handler (called after max retries are reached)."""
        self._failed_request_handler = handler
        return handler

    async def run(self, requests: Sequence[str | BaseRequestData | Request] | None = None) -> FinalStatistics:
        """Run the crawler until all requests are processed."""
        if self._running:
            raise RuntimeError(
                'This crawler instance is already running, you can add more requests to it via `crawler.add_requests()`'
            )

        self._running = True

        if self._has_finished_before:
            await self._statistics.reset()

            if self._use_session_pool:
                await self._session_pool.reset_store()

        if requests is not None:
            await self.add_requests(requests)

        interrupted = False

        def sigint_handler() -> None:
            nonlocal interrupted

            if not interrupted:
                interrupted = True
                self._logger.info('Pausing... Press CTRL+C again to force exit.')

            run_task.cancel()

        run_task = asyncio.create_task(self._run_crawler())

        with suppress(NotImplementedError):  # event loop signal handlers are not supported on Windows
            asyncio.get_running_loop().add_signal_handler(signal.SIGINT, sigint_handler)

        try:
            await run_task
        except CancelledError:
            pass
        finally:
            with suppress(NotImplementedError):
                asyncio.get_running_loop().remove_signal_handler(signal.SIGINT)

        if self._statistics.error_tracker.total > 0:
            self._logger.info(
                'Error analysis:'
                f' total_errors={self._statistics.error_tracker.total}'
                f' unique_errors={self._statistics.error_tracker.unique_error_count}'
            )

        if interrupted:
            self._logger.info(
                f'The crawl was interrupted. To resume, do: CRAWLEE_PURGE_ON_START=0 python {sys.argv[0]}'
            )

        self._running = False
        self._has_finished_before = True

        final_statistics = self._statistics.calculate()
        self._logger.info(f'Final request statistics: {final_statistics}')

        return final_statistics

    async def _run_crawler(self) -> None:
        async with AsyncExitStack() as exit_stack:
            await exit_stack.enter_async_context(self._event_manager)
            await exit_stack.enter_async_context(self._snapshotter)
            await exit_stack.enter_async_context(self._statistics)

            if self._use_session_pool:
                await exit_stack.enter_async_context(self._session_pool)

            for context_manager in self._additional_context_managers:
                await exit_stack.enter_async_context(context_manager)

            await self._pool.run()

    async def add_requests(
        self,
        requests: Sequence[str | BaseRequestData | Request],
        *,
        batch_size: int = 1000,
        wait_time_between_batches: timedelta = timedelta(0),
        wait_for_all_requests_to_be_added: bool = False,
        wait_for_all_requests_to_be_added_timeout: timedelta | None = None,
    ) -> None:
        """Add requests to the underlying request provider in batches.

        Args:
            requests: A list of requests to add to the queue.
            batch_size: The number of requests to add in one batch.
            wait_time_between_batches: Time to wait between adding batches.
            wait_for_all_requests_to_be_added: If True, wait for all requests to be added before returning.
            wait_for_all_requests_to_be_added_timeout: Timeout for waiting for all requests to be added.
        """
        request_provider = await self.get_request_provider()

        await request_provider.add_requests_batched(
            requests=requests,
            batch_size=batch_size,
            wait_time_between_batches=wait_time_between_batches,
            wait_for_all_requests_to_be_added=wait_for_all_requests_to_be_added,
            wait_for_all_requests_to_be_added_timeout=wait_for_all_requests_to_be_added_timeout,
        )

    async def get_data(
        self,
        dataset_id: str | None = None,
        dataset_name: str | None = None,
        **kwargs: Unpack[GetDataKwargs],
    ) -> DatasetItemsListPage:
        """Retrieve data from a dataset.

        This helper method simplifies the process of retrieving data from a dataset. It opens the specified
        dataset and then retrieves the data based on the provided parameters.

        Args:
            dataset_id: The ID of the dataset.
            dataset_name: The name of the dataset.
            kwargs: Keyword arguments to be passed to the dataset's `get_data` method.

        Returns:
            The retrieved data.
        """
        dataset = await Dataset.open(id=dataset_id, name=dataset_name)
        return await dataset.get_data(**kwargs)

    async def export_data(
        self,
        path: str | Path,
        content_type: Literal['json', 'csv'] | None = None,
        dataset_id: str | None = None,
        dataset_name: str | None = None,
    ) -> None:
        """Export data from a dataset.

        This helper method simplifies the process of exporting data from a dataset. It opens the specified
        dataset and then exports the data based on the provided parameters.

        Args:
            path: The destination path
            content_type: The output format
            dataset_id: The ID of the dataset.
            dataset_name: The name of the dataset.
        """
        dataset = await Dataset.open(id=dataset_id, name=dataset_name)
        path = path if isinstance(path, Path) else Path(path)

        if content_type is None:
            content_type = 'csv' if path.suffix == '.csv' else 'json'

        return await dataset.write_to(content_type, path.open('w', newline=''))

    async def _push_data(
        self,
        data: JSONSerializable,
        dataset_id: str | None = None,
        dataset_name: str | None = None,
        **kwargs: Unpack[PushDataKwargs],
    ) -> None:
        """Push data to a dataset.

        This helper method simplifies the process of pushing data to a dataset. It opens the specified
        dataset and then pushes the provided data to it.

        Args:
            data: The data to push to the dataset.
            dataset_id: The ID of the dataset.
            dataset_name: The name of the dataset.
            kwargs: Keyword arguments to be passed to the dataset's `push_data` method.
        """
        dataset = await Dataset.open(id=dataset_id, name=dataset_name)
        await dataset.push_data(data, **kwargs)

    def _should_retry_request(self, crawling_context: BasicCrawlingContext, error: Exception) -> bool:
        if crawling_context.request.no_retry:
            return False

        if isinstance(error, SessionError):
            return ((crawling_context.request.session_rotation_count or 0) + 1) < self._max_session_rotations

        max_request_retries = crawling_context.request.max_retries
        if max_request_retries is None:
            max_request_retries = self._max_request_retries

        return (crawling_context.request.retry_count + 1) < max_request_retries

    async def _check_url_after_redirects(
        self, crawling_context: TCrawlingContext
    ) -> AsyncGenerator[TCrawlingContext, None]:
        """Invoked at the end of the context pipeline to make sure that the `loaded_url` still matches enqueue_strategy.

        This is done to filter out links that redirect outside of the crawled domain.
        """
        if crawling_context.request.loaded_url is not None and not self._check_enqueue_strategy(
            crawling_context.request.enqueue_strategy,
            origin_url=httpx.URL(crawling_context.request.url),
            target_url=httpx.URL(crawling_context.request.loaded_url),
        ):
            raise ContextPipelineInterruptedError(
                f'Skipping URL {crawling_context.request.loaded_url} (redirected from {crawling_context.request.url})'
            )

        yield crawling_context

    def _check_enqueue_strategy(
        self, strategy: EnqueueStrategy, *, target_url: httpx.URL, origin_url: httpx.URL
    ) -> bool:
        """Check if a URL matches the enqueue_strategy."""
        if strategy == EnqueueStrategy.SAME_HOSTNAME:
            return target_url.host == origin_url.host

        if strategy == EnqueueStrategy.SAME_DOMAIN:
            origin_domain = self._tld_extractor.extract_str(origin_url.host).domain
            target_domain = self._tld_extractor.extract_str(target_url.host).domain
            return origin_domain == target_domain

        if strategy == EnqueueStrategy.SAME_ORIGIN:
            return target_url.host == origin_url.host and target_url.scheme == origin_url.scheme

        if strategy == EnqueueStrategy.ALL:
            return True

        assert_never(strategy)

    def _check_url_patterns(
        self,
        target_url: httpx.URL,
        include: Sequence[re.Pattern[Any] | Glob] | None,
        exclude: Sequence[re.Pattern[Any] | Glob] | None,
    ) -> bool:
        """Check if a URL matches configured include/exclude patterns."""
        # If the URL matches any `exclude` pattern, reject it
        for pattern in exclude or ():
            if isinstance(pattern, Glob):
                pattern = pattern.regexp  # noqa: PLW2901

            if pattern.match(str(target_url)) is not None:
                return False

        # If there are no `include` patterns and the URL passed all `exclude` patterns, accept the URL
        if include is None:
            return True

        # If the URL matches any `include` pattern, accept it
        for pattern in include:
            if isinstance(pattern, Glob):
                pattern = pattern.regexp  # noqa: PLW2901

            if pattern.match(str(target_url)) is not None:
                return True

        # The URL does not match any `include` pattern - reject it
        return False

    async def _handle_request_error(self, crawling_context: TCrawlingContext, error: Exception) -> None:
        request_provider = await self.get_request_provider()
        request = crawling_context.request

        if self._should_retry_request(crawling_context, error):
            request.retry_count += 1
            self._statistics.error_tracker.add(error)

            if self._error_handler:
                try:
                    new_request = await self._error_handler(crawling_context, error)
                except Exception as e:
                    raise UserDefinedErrorHandlerError('Exception thrown in user-defined request error handler') from e
                else:
                    if new_request is not None:
                        request = new_request

            await request_provider.reclaim_request(request)
        else:
            await wait_for(
                lambda: request_provider.mark_request_as_handled(crawling_context.request),
                timeout=self._internal_timeout,
                timeout_message='Marking request as handled timed out after '
                f'{self._internal_timeout.total_seconds()} seconds',
                logger=self._logger,
                max_retries=3,
            )
            await self._handle_failed_request(crawling_context, error)
            self._statistics.record_request_processing_failure(request.id or request.unique_key)

    async def _handle_failed_request(self, crawling_context: TCrawlingContext, error: Exception) -> None:
        self._logger.exception('Request failed and reached maximum retries', exc_info=error)
        self._statistics.error_tracker.add(error)

        if self._failed_request_handler:
            try:
                await self._failed_request_handler(crawling_context, error)
            except Exception as e:
                raise UserDefinedErrorHandlerError('Exception thrown in user-defined failed request handler') from e

    def _prepare_send_request_function(
        self,
        session: Session | None,
        proxy_info: ProxyInfo | None,
    ) -> SendRequestFunction:
        async def send_request(
            url: str,
            *,
            method: str = 'get',
            headers: dict[str, str] | None = None,
        ) -> HttpResponse:
            return await self._http_client.send_request(
                url,
                method=method,
                headers=httpx.Headers(headers),
                session=session,
                proxy_info=proxy_info,
            )

        return send_request

    async def _commit_request_handler_result(
        self, context: BasicCrawlingContext, result: RequestHandlerRunResult
    ) -> None:
        request_provider = await self.get_request_provider()
        origin = httpx.URL(context.request.loaded_url or context.request.url)

        for call in result.add_requests_calls:
            requests = list[BaseRequestData]()

            for request in call['requests']:
                if (limit := call.get('limit')) is not None and len(requests) >= limit:
                    break

                request_model = request if isinstance(request, BaseRequestData) else BaseRequestData.from_url(request)
                destination = httpx.URL(request_model.url)
                if destination.is_relative_url:
                    base_url = httpx.URL(call.get('base_url', origin))
                    destination = base_url.join(destination)
                    request_model.url = str(destination)

                if self._check_enqueue_strategy(
                    call.get('strategy', EnqueueStrategy.ALL), target_url=destination, origin_url=origin
                ) and self._check_url_patterns(destination, call.get('include', None), call.get('exclude', None)):
                    requests.append(request_model)

            await request_provider.add_requests_batched(requests)

    async def __is_finished_function(self) -> bool:
        request_provider = await self.get_request_provider()
        is_finished = await request_provider.is_finished()

        if self._max_requests_count_exceeded:
            self._logger.info(
                f'The crawler has reached its limit of {self._max_requests_per_crawl} requests per crawl. '
                f'All ongoing requests have now completed. Total requests processed: '
                f'{self._statistics.state.requests_finished}. The crawler will now shut down.'
            )
            self._logger.info(f'is_finished: {is_finished}')
            return True

        return is_finished

    async def __is_task_ready_function(self) -> bool:
        if self._max_requests_count_exceeded:
            self._logger.info(
                f'The crawler has reached its limit of {self._max_requests_per_crawl} requests per crawl. '
                f'The crawler will soon shut down. Ongoing requests will be allowed to complete.'
            )
            return False

        request_provider = await self.get_request_provider()
        return not await request_provider.is_empty()

    async def __run_task_function(self) -> None:
        request_provider = await self.get_request_provider()

        request = await wait_for(
            lambda: request_provider.fetch_next_request(),
            timeout=self._internal_timeout,
            timeout_message=f'Fetching next request failed after {self._internal_timeout.total_seconds()} seconds',
            logger=self._logger,
            max_retries=3,
        )

        if request is None:
            return

        session = await self._get_session()
        proxy_info = await self._get_proxy_info(request, session)
        result = RequestHandlerRunResult()

        crawling_context = BasicCrawlingContext(
            request=request,
            session=session,
            proxy_info=proxy_info,
            send_request=self._prepare_send_request_function(session, proxy_info),
            add_requests=result.add_requests,
            push_data=self._push_data,
            log=self._logger,
        )

        statistics_id = request.id or request.unique_key
        self._statistics.record_request_processing_start(statistics_id)

        try:
            request.state = RequestState.REQUEST_HANDLER

            await wait_for(
                lambda: self.__run_request_handler(crawling_context),
                timeout=self._request_handler_timeout,
                timeout_message='Request handler timed out after '
                f'{self._request_handler_timeout.total_seconds()} seconds',
                logger=self._logger,
            )

            await self._commit_request_handler_result(crawling_context, result)

            await wait_for(
                lambda: request_provider.mark_request_as_handled(crawling_context.request),
                timeout=self._internal_timeout,
                timeout_message='Marking request as handled timed out after '
                f'{self._internal_timeout.total_seconds()} seconds',
                logger=self._logger,
                max_retries=3,
            )

            request.state = RequestState.DONE

            if crawling_context.session:
                crawling_context.session.mark_good()

            self._statistics.record_request_processing_finish(statistics_id)
        except RequestHandlerError as primary_error:
            primary_error = cast(
                RequestHandlerError[TCrawlingContext], primary_error
            )  # valid thanks to ContextPipeline

            try:
                request.state = RequestState.ERROR_HANDLER

                await wait_for(
                    partial(
                        self._handle_request_error, primary_error.crawling_context, primary_error.wrapped_exception
                    ),
                    timeout=self._internal_timeout,
                    timeout_message='Handling request failure timed out after '
                    f'{self._internal_timeout.total_seconds()} seconds',
                    logger=self._logger,
                )

                request.state = RequestState.DONE
            except UserDefinedErrorHandlerError:
                request.state = RequestState.ERROR
                raise
            except Exception as secondary_error:
                self._logger.exception(
                    'An exception occurred during handling of failed request. This places the crawler '
                    'and its underlying storages into an unknown state and crawling will be terminated.',
                    exc_info=secondary_error,
                )
                request.state = RequestState.ERROR
                raise

            if crawling_context.session:
                crawling_context.session.mark_bad()
        except SessionError as session_error:
            if not crawling_context.session:
                raise RuntimeError('SessionError raised in a crawling context without a session') from session_error

            if self._should_retry_request(crawling_context, session_error):
                self._logger.warning('Encountered a session error, rotating session and retrying')

                crawling_context.session.retire()

                if crawling_context.request.session_rotation_count is None:
                    crawling_context.request.session_rotation_count = 0
                crawling_context.request.session_rotation_count += 1

                await request_provider.reclaim_request(request)
                self._statistics.error_tracker_retry.add(session_error)
            else:
                self._logger.exception('Request failed and reached maximum retries', exc_info=session_error)

                await wait_for(
                    lambda: request_provider.mark_request_as_handled(crawling_context.request),
                    timeout=self._internal_timeout,
                    timeout_message='Marking request as handled timed out after '
                    f'{self._internal_timeout.total_seconds()} seconds',
                    logger=self._logger,
                    max_retries=3,
                )

                self._statistics.record_request_processing_failure(statistics_id)
                self._statistics.error_tracker.add(session_error)
        except ContextPipelineInterruptedError as interrupted_error:
            self._logger.debug('The context pipeline was interrupted', exc_info=interrupted_error)

            await wait_for(
                lambda: request_provider.mark_request_as_handled(crawling_context.request),
                timeout=self._internal_timeout,
                timeout_message='Marking request as handled timed out after '
                f'{self._internal_timeout.total_seconds()} seconds',
                logger=self._logger,
                max_retries=3,
            )
        except ContextPipelineInitializationError as initialization_error:
            if self._should_retry_request(crawling_context, initialization_error):
                self._logger.debug(
                    'An exception occurred during the initialization of crawling context, a retry is in order',
                    exc_info=initialization_error,
                )

                request = crawling_context.request
                request.retry_count += 1
                request.state = RequestState.DONE
                await request_provider.reclaim_request(request)
            else:
                self._logger.exception('Request failed and reached maximum retries', exc_info=initialization_error)

                await wait_for(
                    lambda: request_provider.mark_request_as_handled(crawling_context.request),
                    timeout=self._internal_timeout,
                    timeout_message='Marking request as handled timed out after '
                    f'{self._internal_timeout.total_seconds()} seconds',
                    logger=self._logger,
                    max_retries=3,
                )

            if crawling_context.session:
                crawling_context.session.mark_bad()
        except Exception as internal_error:
            self._logger.exception(
                'An exception occurred during handling of a request. This places the crawler '
                'and its underlying storages into an unknown state and crawling will be terminated.',
                exc_info=internal_error,
            )
            raise

    async def __run_request_handler(self, crawling_context: BasicCrawlingContext) -> None:
        await self._context_pipeline(crawling_context, self.router)



from __future__ import annotations

from collections import deque
from datetime import timedelta
from typing import TYPE_CHECKING

from typing_extensions import override

from crawlee.storages.request_provider import RequestProvider

if TYPE_CHECKING:
    from collections.abc import Sequence

    from crawlee.models import BaseRequestData, Request


class RequestList(RequestProvider):
    """Represents a (potentially very large) list of URLs to crawl."""

    def __init__(
        self,
        requests: Sequence[str | BaseRequestData | Request] | None = None,
        name: str | None = None,
    ) -> None:
        """Initialize the RequestList.

        Args:
            requests: the URLs (or crawling requests) to crawl
            name: a name of the request list
        """
        self._name = name or ''
        self._handled_count = 0

        self._requests = deque(self._transform_requests(requests or []))
        self._in_progress = set[str]()

    @property
    @override
    def name(self) -> str:
        return self._name

    @override
    async def get_total_count(self) -> int:
        return len(self._requests)

    @override
    async def is_empty(self) -> bool:
        return len(self._requests) == 0

    @override
    async def is_finished(self) -> bool:
        return await self.is_empty() and len(self._in_progress) == 0

    @override
    async def drop(self) -> None:
        self._requests.clear()

    @override
    async def fetch_next_request(self) -> Request | None:
        try:
            request = self._requests.popleft()
        except IndexError:
            return None
        else:
            self._in_progress.add(request.id)
            return request

    @override
    async def reclaim_request(self, request: Request, *, forefront: bool = False) -> None:
        if forefront:
            self._requests.appendleft(request)
        else:
            self._requests.append(request)

        self._in_progress.remove(request.id)

    @override
    async def mark_request_as_handled(self, request: Request) -> None:
        self._handled_count += 1
        self._in_progress.remove(request.id)

    @override
    async def get_handled_count(self) -> int:
        return self._handled_count

    @override
    async def add_requests_batched(
        self,
        requests: Sequence[str | BaseRequestData | Request],
        *,
        batch_size: int = 1000,
        wait_time_between_batches: timedelta = timedelta(seconds=1),
        wait_for_all_requests_to_be_added: bool = False,
        wait_for_all_requests_to_be_added_timeout: timedelta | None = None,
    ) -> None:
        transformed_requests = self._transform_requests(requests)
        self._requests.extend(transformed_requests)
